from collections import OrderedDict, defaultdict
from dataclasses import dataclass

import numpy as np
import pytest

from astropy.io import fits
from astropy.utils import metadata
from astropy.utils.metadata import (
    MergeConflictError,
    MetaData,
    common_dtype,
    enable_merge_strategies,
    merge,
)


class OrderedDictSubclass(OrderedDict):
    pass


class MetaBaseTest:
    def test_none(self):
        d = self.test_class(*self.args)
        assert isinstance(d.meta, dict)
        assert len(d.meta) == 0

    @pytest.mark.parametrize(
        "meta",
        ([{"a": 1}, OrderedDict([("a", 1)]), OrderedDictSubclass([("a", 1)])]),
    )
    def test_mapping_init(self, meta):
        d = self.test_class(*self.args, meta=meta)
        assert type(d.meta) == type(meta)
        assert d.meta["a"] == 1

    @pytest.mark.parametrize("meta", (["ceci n'est pas un meta", 1.2, [1, 2, 3]]))
    def test_non_mapping_init(self, meta):
        with pytest.raises(TypeError):
            self.test_class(*self.args, meta=meta)

    @pytest.mark.parametrize(
        "meta",
        ([{"a": 1}, OrderedDict([("a", 1)]), OrderedDictSubclass([("a", 1)])]),
    )
    def test_mapping_set(self, meta):
        d = self.test_class(*self.args, meta=meta)
        assert type(d.meta) == type(meta)
        assert d.meta["a"] == 1

    @pytest.mark.parametrize("meta", (["ceci n'est pas un meta", 1.2, [1, 2, 3]]))
    def test_non_mapping_set(self, meta):
        with pytest.raises(TypeError):
            d = self.test_class(*self.args, meta=meta)

    def test_meta_fits_header(self):
        header = fits.header.Header()
        header.set("observer", "Edwin Hubble")
        header.set("exptime", "3600")

        d = self.test_class(*self.args, meta=header)

        assert d.meta["OBSERVER"] == "Edwin Hubble"


class ExampleData:
    meta = MetaData()

    def __init__(self, meta=None):
        self.meta = meta


class TestMetaExampleData(MetaBaseTest):
    test_class = ExampleData
    args = ()


@dataclass
class ExampleDataclass:
    meta: MetaData = MetaData()  # noqa: RUF009


class TestMetaExampleDataclass(MetaBaseTest):
    test_class = ExampleDataclass
    args = ()


@dataclass(frozen=True)
class ExampleFrozenDataclass:
    meta: MetaData = MetaData()  # noqa: RUF009


class TestMetaExampleFrozenDataclass(MetaBaseTest):
    test_class = ExampleFrozenDataclass
    args = ()


def test_metadata_default():
    data = ExampleDataclass()
    data.meta["a"] = 1
    assert isinstance(data.meta, OrderedDict)


def test_metadata_default_factory():
    """Test the default_factory argument to MetaData."""

    class ExampleData:
        meta = MetaData(default_factory=defaultdict)

        def __init__(self, meta=None):
            self.meta = meta

    data = ExampleData()
    assert isinstance(data.meta, defaultdict)
    assert len(data.meta) == 0


def test_metadata_merging_conflict_exception():
    """Regression test for issue #3294.

    Ensure that an exception is raised when a metadata conflict exists
    and ``metadata_conflicts='error'`` has been set.
    """
    data1 = ExampleData()
    data2 = ExampleData()
    data1.meta["somekey"] = {"x": 1, "y": 1}
    data2.meta["somekey"] = {"x": 1, "y": 999}
    with pytest.raises(MergeConflictError):
        merge(data1.meta, data2.meta, metadata_conflicts="error")


def test_metadata_merging():
    # Recursive merge
    meta1 = {
        "k1": {
            "k1": [1, 2],
            "k2": 2,
        },
        "k2": 2,
        "k4": (1, 2),
    }
    meta2 = {
        "k1": {"k1": [3]},
        "k3": 3,
        "k4": (3,),
    }
    out = merge(meta1, meta2, metadata_conflicts="error")
    assert out == {
        "k1": {
            "k2": 2,
            "k1": [1, 2, 3],
        },
        "k2": 2,
        "k3": 3,
        "k4": (1, 2, 3),
    }

    # Merge two ndarrays
    meta1 = {"k1": np.array([1, 2])}
    meta2 = {"k1": np.array([3])}
    out = merge(meta1, meta2, metadata_conflicts="error")
    assert np.all(out["k1"] == np.array([1, 2, 3]))

    # Merge list and np.ndarray
    meta1 = {"k1": [1, 2]}
    meta2 = {"k1": np.array([3])}
    assert np.all(out["k1"] == np.array([1, 2, 3]))

    # Can't merge two scalar types
    meta1 = {"k1": 1}
    meta2 = {"k1": 2}
    with pytest.raises(MergeConflictError):
        merge(meta1, meta2, metadata_conflicts="error")

    # Conflicting shape
    meta1 = {"k1": np.array([1, 2])}
    meta2 = {"k1": np.array([[3]])}
    with pytest.raises(MergeConflictError):
        merge(meta1, meta2, metadata_conflicts="error")

    # Conflicting array type
    meta1 = {"k1": np.array([1, 2])}
    meta2 = {"k1": np.array(["3"])}
    with pytest.raises(MergeConflictError):
        merge(meta1, meta2, metadata_conflicts="error")

    # Conflicting array type with 'silent' merging
    meta1 = {"k1": np.array([1, 2])}
    meta2 = {"k1": np.array(["3"])}
    out = merge(meta1, meta2, metadata_conflicts="silent")
    assert np.all(out["k1"] == np.array(["3"]))


def test_metadata_merging_new_strategy():
    original_merge_strategies = list(metadata.MERGE_STRATEGIES)

    class MergeNumbersAsList(metadata.MergeStrategy):
        """
        Scalar float or int values are joined in a list.
        """

        types = ((int, float), (int, float))

        @classmethod
        def merge(cls, left, right):
            return [left, right]

    class MergeConcatStrings(metadata.MergePlus):
        """
        Scalar string values are concatenated
        """

        types = (str, str)
        enabled = False

    # Normally can't merge two scalar types
    meta1 = {"k1": 1, "k2": "a"}
    meta2 = {"k1": 2, "k2": "b"}

    # Enable new merge strategy
    with enable_merge_strategies(MergeNumbersAsList, MergeConcatStrings):
        assert MergeNumbersAsList.enabled
        assert MergeConcatStrings.enabled
        out = merge(meta1, meta2, metadata_conflicts="error")
    assert out["k1"] == [1, 2]
    assert out["k2"] == "ab"
    assert not MergeNumbersAsList.enabled
    assert not MergeConcatStrings.enabled

    # Confirm the default enabled=False behavior
    with pytest.raises(MergeConflictError):
        merge(meta1, meta2, metadata_conflicts="error")

    # Enable all MergeStrategy subclasses
    with enable_merge_strategies(metadata.MergeStrategy):
        assert MergeNumbersAsList.enabled
        assert MergeConcatStrings.enabled
        out = merge(meta1, meta2, metadata_conflicts="error")
    assert out["k1"] == [1, 2]
    assert out["k2"] == "ab"
    assert not MergeNumbersAsList.enabled
    assert not MergeConcatStrings.enabled

    metadata.MERGE_STRATEGIES = original_merge_strategies


def test_common_dtype_string():
    u3 = np.array(["123"])
    u4 = np.array(["1234"])
    b3 = np.array([b"123"])
    b5 = np.array([b"12345"])
    assert common_dtype([u3, u4]).endswith("U4")
    assert common_dtype([b5, u4]).endswith("U5")
    assert common_dtype([b3, b5]).endswith("S5")


def test_common_dtype_basic():
    i8 = np.array(1, dtype=np.int64)
    f8 = np.array(1, dtype=np.float64)
    u3 = np.array("123")

    with pytest.raises(MergeConflictError):
        common_dtype([i8, u3])

    assert common_dtype([i8, i8]).endswith("i8")
    assert common_dtype([i8, f8]).endswith("f8")



"""Scipy compatibility."""

from typing import Any, Never

from astropy.utils.compat.optional_deps import HAS_SCIPY

__all__ = ("ellipkinc", "hyp2f1", "quad")

if HAS_SCIPY:
    from scipy.integrate import quad
    from scipy.special import ellipkinc, hyp2f1

else:

    def quad(*args: Any, **kwargs: Any) -> Never:
        raise ModuleNotFoundError("No module named 'scipy.integrate'")

    def ellipkinc(*args: Any, **kwargs: Any) -> Never:
        raise ModuleNotFoundError("No module named 'scipy.special'")

    def hyp2f1(*args: Any, **kwargs: Any) -> Never:
        raise ModuleNotFoundError("No module named 'scipy.special'")



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Traits for building ``astropy`` :class:`~astropy.cosmology.Cosmology` classes."""

__all__ = [
    "BaryonComponent",
    "CriticalDensity",
    "CurvatureComponent",
    "DarkEnergyComponent",
    "DarkMatterComponent",
    "HubbleParameter",
    "MatterComponent",
    "PhotonComponent",
    "ScaleFactor",
    "TemperatureCMB",
    "TotalComponent",
]

from ._src.traits import (
    BaryonComponent,
    CriticalDensity,
    CurvatureComponent,
    DarkEnergyComponent,
    DarkMatterComponent,
    HubbleParameter,
    MatterComponent,
    PhotonComponent,
    ScaleFactor,
    TemperatureCMB,
    TotalComponent,
)



# Licensed under a 3-clause BSD style license - see LICENSE.rst


from io import StringIO
from shutil import get_terminal_size

import numpy as np
import pytest

from astropy import conf, table
from astropy import units as u
from astropy.io import ascii
from astropy.table import Column, QTable, Table
from astropy.table.table_helpers import simple_table
from astropy.utils.exceptions import AstropyDeprecationWarning

BIG_WIDE_ARR = np.arange(2000, dtype=np.float64).reshape(100, 20)
SMALL_ARR = np.arange(18, dtype=np.int64).reshape(6, 3)


@pytest.mark.usefixtures("table_type")
class TestMultiD:
    def test_multidim(self, table_type):
        """Test printing with multidimensional column"""
        arr = [
            np.array([[1, 2], [10, 20]], dtype=np.int64),
            np.array([[3, 4], [30, 40]], dtype=np.int64),
            np.array([[5, 6], [50, 60]], dtype=np.int64),
        ]
        t = table_type(arr)
        lines = t.pformat(show_dtype=True)
        assert lines == [
            "  col0     col1     col2  ",
            "int64[2] int64[2] int64[2]",
            "-------- -------- --------",
            "  1 .. 2   3 .. 4   5 .. 6",
            "10 .. 20 30 .. 40 50 .. 60",
        ]

        lines = t.pformat(html=True, show_dtype=True)
        assert lines == [
            f'<table id="table{id(t)}">',
            "<thead><tr><th>col0</th><th>col1</th><th>col2</th></tr></thead>",
            "<thead><tr><th>int64[2]</th><th>int64[2]</th><th>int64[2]</th></tr></thead>",
            "<tr><td>1 .. 2</td><td>3 .. 4</td><td>5 .. 6</td></tr>",
            "<tr><td>10 .. 20</td><td>30 .. 40</td><td>50 .. 60</td></tr>",
            "</table>",
        ]
        nbclass = table.conf.default_notebook_table_class
        masked = "masked=True " if t.masked else ""
        assert t._repr_html_().splitlines() == [
            f"<div><i>{table_type.__name__} {masked}length=2</i>",
            f'<table id="table{id(t)}" class="{nbclass}">',
            "<thead><tr><th>col0</th><th>col1</th><th>col2</th></tr></thead>",
            "<thead><tr><th>int64[2]</th><th>int64[2]</th><th>int64[2]</th></tr></thead>",
            "<tr><td>1 .. 2</td><td>3 .. 4</td><td>5 .. 6</td></tr>",
            "<tr><td>10 .. 20</td><td>30 .. 40</td><td>50 .. 60</td></tr>",
            "</table></div>",
        ]

        t = table_type([arr])
        lines = t.pformat(show_dtype=True)
        assert lines == [
            "   col0   ",
            "int64[2,2]",
            "----------",
            "   1 .. 20",
            "   3 .. 40",
            "   5 .. 60",
        ]

    def test_fake_multidim(self, table_type):
        """Test printing with 'fake' multidimensional column"""
        arr = [
            np.array([[(1,)], [(10,)]], dtype=np.int64),
            np.array([[(3,)], [(30,)]], dtype=np.int64),
            np.array([[(5,)], [(50,)]], dtype=np.int64),
        ]
        t = table_type(arr)
        lines = t.pformat(show_dtype=True)
        assert lines == [
            "   col0       col1       col2   ",
            "int64[1,1] int64[1,1] int64[1,1]",
            "---------- ---------- ----------",
            "         1          3          5",
            "        10         30         50",
        ]

        lines = t.pformat(html=True, show_dtype=True)
        assert lines == [
            f'<table id="table{id(t)}">',
            "<thead><tr><th>col0</th><th>col1</th><th>col2</th></tr></thead>",
            "<thead><tr><th>int64[1,1]</th><th>int64[1,1]</th><th>int64[1,1]</th></tr></thead>",
            "<tr><td>1</td><td>3</td><td>5</td></tr>",
            "<tr><td>10</td><td>30</td><td>50</td></tr>",
            "</table>",
        ]
        nbclass = table.conf.default_notebook_table_class
        masked = "masked=True " if t.masked else ""
        assert t._repr_html_().splitlines() == [
            f"<div><i>{table_type.__name__} {masked}length=2</i>",
            f'<table id="table{id(t)}" class="{nbclass}">',
            "<thead><tr><th>col0</th><th>col1</th><th>col2</th></tr></thead>",
            "<thead><tr><th>int64[1,1]</th><th>int64[1,1]</th><th>int64[1,1]</th></tr></thead>",
            "<tr><td>1</td><td>3</td><td>5</td></tr>",
            "<tr><td>10</td><td>30</td><td>50</td></tr>",
            "</table></div>",
        ]

        t = table_type([arr])
        lines = t.pformat(show_dtype=True)
        assert lines == [
            "    col0    ",
            "int64[2,1,1]",
            "------------",
            "     1 .. 10",
            "     3 .. 30",
            "     5 .. 50",
        ]


def test_html_escaping():
    t = table.Table([('<script>alert("gotcha");</script>', 2, 3)])
    nbclass = table.conf.default_notebook_table_class
    assert t._repr_html_().splitlines() == [
        "<div><i>Table length=3</i>",
        f'<table id="table{id(t)}" class="{nbclass}">',
        "<thead><tr><th>col0</th></tr></thead>",
        "<thead><tr><th>str33</th></tr></thead>",
        "<tr><td>&lt;script&gt;alert(&quot;gotcha&quot;);&lt;/script&gt;</td></tr>",
        "<tr><td>2</td></tr>",
        "<tr><td>3</td></tr>",
        "</table></div>",
    ]


@pytest.mark.usefixtures("table_type")
class TestPprint:
    def _setup(self, table_type):
        self.tb = table_type(BIG_WIDE_ARR)
        self.tb["col0"].format = "e"
        self.tb["col1"].format = ".6f"

        self.tb["col0"].unit = "km**2"
        self.tb["col19"].unit = "kg s m**-2"
        self.ts = table_type(SMALL_ARR)

    def test_empty_table(self, table_type):
        t = table_type()
        lines = t.pformat()
        assert lines == ["<No columns>"]
        c = repr(t)
        masked = "masked=True " if t.masked else ""
        assert c.splitlines() == [
            f"<{table_type.__name__} {masked}length=0>",
            "<No columns>",
        ]

    def test_format0(self, table_type):
        """Try getting screen size but fail to defaults because testing doesn't
        have access to screen (fcntl.ioctl fails).
        """
        self._setup(table_type)
        arr = np.arange(4000, dtype=np.float64).reshape(100, 40)
        with conf.set_temp("max_width", None), conf.set_temp("max_lines", None):
            lines = table_type(arr).pformat(max_lines=None, max_width=None)
        width, nlines = get_terminal_size()
        assert len(lines) == nlines
        for line in lines[:-1]:  # skip last "Length = .. rows" line
            assert width - 10 < len(line) <= width

    def test_format1(self, table_type):
        """Basic test of formatting, unit header row included"""
        self._setup(table_type)
        lines = self.tb.pformat(max_lines=8, max_width=40)
        assert lines == [
            "    col0         col1    ...   col19  ",
            "    km2                  ... kg s / m2",
            "------------ ----------- ... ---------",
            "0.000000e+00    1.000000 ...      19.0",
            "         ...         ... ...       ...",
            "1.960000e+03 1961.000000 ...    1979.0",
            "1.980000e+03 1981.000000 ...    1999.0",
            "Length = 100 rows",
        ]

    def test_format2(self, table_type):
        """Basic test of formatting, unit header row excluded"""
        self._setup(table_type)
        lines = self.tb.pformat(max_lines=8, max_width=40, show_unit=False)
        assert lines == [
            "    col0         col1    ... col19 ",
            "------------ ----------- ... ------",
            "0.000000e+00    1.000000 ...   19.0",
            "2.000000e+01   21.000000 ...   39.0",
            "         ...         ... ...    ...",
            "1.960000e+03 1961.000000 ... 1979.0",
            "1.980000e+03 1981.000000 ... 1999.0",
            "Length = 100 rows",
        ]

    def test_format3(self, table_type):
        """Include the unit header row"""
        self._setup(table_type)
        lines = self.tb.pformat(max_lines=8, max_width=40, show_unit=True)

        assert lines == [
            "    col0         col1    ...   col19  ",
            "    km2                  ... kg s / m2",
            "------------ ----------- ... ---------",
            "0.000000e+00    1.000000 ...      19.0",
            "         ...         ... ...       ...",
            "1.960000e+03 1961.000000 ...    1979.0",
            "1.980000e+03 1981.000000 ...    1999.0",
            "Length = 100 rows",
        ]

    def test_format4(self, table_type):
        """Do not include the name header row"""
        self._setup(table_type)
        lines = self.tb.pformat(max_lines=8, max_width=40, show_name=False)
        assert lines == [
            "    km2                  ... kg s / m2",
            "------------ ----------- ... ---------",
            "0.000000e+00    1.000000 ...      19.0",
            "2.000000e+01   21.000000 ...      39.0",
            "         ...         ... ...       ...",
            "1.960000e+03 1961.000000 ...    1979.0",
            "1.980000e+03 1981.000000 ...    1999.0",
            "Length = 100 rows",
        ]

    def test_noclip(self, table_type):
        """Basic table print"""
        self._setup(table_type)
        lines = self.ts.pformat(max_lines=-1, max_width=-1)
        assert lines == [
            "col0 col1 col2",
            "---- ---- ----",
            "   0    1    2",
            "   3    4    5",
            "   6    7    8",
            "   9   10   11",
            "  12   13   14",
            "  15   16   17",
        ]

    def test_clip1(self, table_type):
        """max lines below hard limit of 8"""
        self._setup(table_type)
        lines = self.ts.pformat(max_lines=3, max_width=-1)
        assert lines == [
            "col0 col1 col2",
            "---- ---- ----",
            "   0    1    2",
            "   3    4    5",
            "   6    7    8",
            "   9   10   11",
            "  12   13   14",
            "  15   16   17",
        ]

    def test_clip2(self, table_type):
        """max lines below hard limit of 8 and output longer than 8"""
        self._setup(table_type)
        lines = self.ts.pformat(
            max_lines=3, max_width=-1, show_unit=True, show_dtype=True
        )
        assert lines == [
            " col0  col1  col2",
            "                 ",
            "int64 int64 int64",
            "----- ----- -----",
            "    0     1     2",
            "  ...   ...   ...",
            "   15    16    17",
            "Length = 6 rows",
        ]

    def test_clip3(self, table_type):
        """Max lines below hard limit of 8 and max width below hard limit
        of 10
        """
        self._setup(table_type)
        lines = self.ts.pformat(max_lines=3, max_width=1, show_unit=True)
        assert lines == [
            "col0 ...",
            "     ...",
            "---- ...",
            "   0 ...",
            " ... ...",
            "  12 ...",
            "  15 ...",
            "Length = 6 rows",
        ]

    def test_clip4(self, table_type):
        """Test a range of max_lines"""
        self._setup(table_type)
        for max_lines in (0, 1, 4, 5, 6, 7, 8, 100, 101, 102, 103, 104, 130):
            lines = self.tb.pformat(max_lines=max_lines, show_unit=False)
            assert len(lines) == max(8, min(102, max_lines))

    def test_pformat_all(self, table_type):
        """Test that all rows are printed by default"""
        self._setup(table_type)
        with pytest.warns(
            AstropyDeprecationWarning,
            match=(
                r"The pformat_all function is deprecated "
                r"and may be removed in a future version\."
            ),
        ):
            lines = self.tb.pformat_all()
        # +3 accounts for the three header lines in this  table
        assert len(lines) == BIG_WIDE_ARR.shape[0] + 3

    def test_pprint_all(self, table_type, capsys):
        """Test that all rows are printed by default"""
        self._setup(table_type)
        self.tb.pprint_all()
        (out, err) = capsys.readouterr()
        # +3 accounts for the three header lines in this  table
        assert len(out.splitlines()) == BIG_WIDE_ARR.shape[0] + 3


class TestPprintColumn:
    @pytest.mark.parametrize(
        "scalar, exp",
        [
            (
                1,
                [
                    "None",
                    "----",
                    "   1",
                ],
            ),
            (
                u.Quantity(0.6, "eV"),
                [
                    "None",
                    "----",
                    " 0.6",
                ],
            ),
        ],
    )
    def test_pprint_scalar(self, scalar, exp):
        # see https://github.com/astropy/astropy/issues/12584
        c = Column(scalar)

        # Make sure pprint() does not raise an exception
        c.pprint()

        # Check actual output
        out = c.pformat()
        assert out == exp


@pytest.mark.usefixtures("table_type")
class TestFormat:
    def test_column_format(self, table_type):
        t = table_type([[1, 2], [3, 4]], names=("a", "b"))
        # default (format=None)
        assert str(t["a"]) == " a \n---\n  1\n  2"

        # just a plain format string
        t["a"].format = "5.2f"
        assert str(t["a"]) == "  a  \n-----\n 1.00\n 2.00"

        #  Old-style that is almost new-style
        t["a"].format = "{ %4.2f }"
        assert str(t["a"]) == "   a    \n--------\n{ 1.00 }\n{ 2.00 }"

        #  New-style that is almost old-style
        t["a"].format = "%{0:}"
        assert str(t["a"]) == " a \n---\n %1\n %2"

        #  New-style with extra spaces
        t["a"].format = " {0:05d} "
        assert str(t["a"]) == "   a   \n-------\n 00001 \n 00002 "

        #  New-style has precedence
        t["a"].format = "%4.2f {0:}"
        assert str(t["a"]) == "   a   \n-------\n%4.2f 1\n%4.2f 2"

        #  Invalid format spec
        with pytest.raises(ValueError):
            t["a"].format = "fail"
        assert t["a"].format == "%4.2f {0:}"  # format did not change

    def test_column_format_with_threshold(self, table_type):
        from astropy import conf

        with conf.set_temp("max_lines", 8):
            t = table_type([np.arange(20)], names=["a"])
            t["a"].format = "%{0:}"
            assert str(t["a"]).splitlines() == [
                " a ",
                "---",
                " %0",
                " %1",
                "...",
                "%18",
                "%19",
                "Length = 20 rows",
            ]
            t["a"].format = "{ %4.2f }"
            assert str(t["a"]).splitlines() == [
                "    a    ",
                "---------",
                " { 0.00 }",
                " { 1.00 }",
                "      ...",
                "{ 18.00 }",
                "{ 19.00 }",
                "Length = 20 rows",
            ]

    def test_column_format_func(self, table_type):
        # run most of functions twice
        # 1) astropy.table.pprint._format_funcs gets populated
        # 2) astropy.table.pprint._format_funcs gets used

        t = table_type([[1.0, 2.0], [3, 4]], names=("a", "b"))

        # mathematical function
        t["a"].format = lambda x: str(x * 3.0)
        assert str(t["a"]) == " a \n---\n3.0\n6.0"
        assert str(t["a"]) == " a \n---\n3.0\n6.0"

    def test_column_format_callable(self, table_type):
        # run most of functions twice
        # 1) astropy.table.pprint._format_funcs gets populated
        # 2) astropy.table.pprint._format_funcs gets used

        t = table_type([[1.0, 2.0], [3, 4]], names=("a", "b"))

        # mathematical function
        class format:
            def __call__(self, x):
                return str(x * 3.0)

        t["a"].format = format()
        assert str(t["a"]) == " a \n---\n3.0\n6.0"
        assert str(t["a"]) == " a \n---\n3.0\n6.0"

    def test_column_format_func_wrong_number_args(self, table_type):
        t = table_type([[1.0, 2.0], [3, 4]], names=("a", "b"))

        # function that expects wrong number of arguments
        def func(a, b):
            pass

        with pytest.raises(ValueError):
            t["a"].format = func

    def test_column_format_func_multiD(self, table_type):
        arr = [np.array([[1, 2], [10, 20]], dtype="i8")]
        t = table_type(arr, names=["a"])

        # mathematical function
        t["a"].format = lambda x: str(x * 3.0)
        outstr = [
            "     a      ",
            "------------",
            "  3.0 .. 6.0",
            "30.0 .. 60.0",
        ]
        assert str(t["a"]).splitlines() == outstr

    def test_column_format_func_not_str(self, table_type):
        t = table_type([[1.0, 2.0], [3, 4]], names=("a", "b"))

        # mathematical function
        with pytest.raises(ValueError):
            t["a"].format = lambda x: x * 3

    def test_column_alignment(self, table_type):
        t = table_type(
            [[1], [2], [3], [4]],
            names=("long title a", "long title b", "long title c", "long title d"),
        )
        t["long title a"].format = "<"
        t["long title b"].format = "^"
        t["long title c"].format = ">"
        t["long title d"].format = "0="
        assert str(t["long title a"]) == "long title a\n------------\n1           "
        assert str(t["long title b"]) == "long title b\n------------\n     2      "
        assert str(t["long title c"]) == "long title c\n------------\n           3"
        assert str(t["long title d"]) == "long title d\n------------\n000000000004"


class TestFormatWithMaskedElements:
    def test_column_format(self):
        t = Table([[1, 2, 3], [3, 4, 5]], names=("a", "b"), masked=True)
        t["a"].mask = [True, False, True]
        # default (format=None)
        assert str(t["a"]) == " a \n---\n --\n  2\n --"

        # just a plain format string
        t["a"].format = "5.2f"
        assert str(t["a"]) == "  a  \n-----\n   --\n 2.00\n   --"

        #  Old-style that is almost new-style
        t["a"].format = "{ %4.2f }"
        assert str(t["a"]) == "   a    \n--------\n      --\n{ 2.00 }\n      --"

        #  New-style that is almost old-style
        t["a"].format = "%{0:}"
        assert str(t["a"]) == " a \n---\n --\n %2\n --"

        #  New-style with extra spaces
        t["a"].format = " {0:05d} "
        assert str(t["a"]) == "   a   \n-------\n     --\n 00002 \n     --"

        #  New-style has precedence
        t["a"].format = "%4.2f {0:}"
        assert str(t["a"]) == "   a   \n-------\n     --\n%4.2f 2\n     --"

    def test_column_format_with_threshold_masked_table(self):
        from astropy import conf

        with conf.set_temp("max_lines", 8):
            t = Table([np.arange(20)], names=["a"], masked=True)
            t["a"].format = "%{0:}"
            t["a"].mask[0] = True
            t["a"].mask[-1] = True
            assert str(t["a"]).splitlines() == [
                " a ",
                "---",
                " --",
                " %1",
                "...",
                "%18",
                " --",
                "Length = 20 rows",
            ]
            t["a"].format = "{ %4.2f }"
            assert str(t["a"]).splitlines() == [
                "    a    ",
                "---------",
                "       --",
                " { 1.00 }",
                "      ...",
                "{ 18.00 }",
                "       --",
                "Length = 20 rows",
            ]

    def test_column_format_func(self):
        # run most of functions twice
        # 1) astropy.table.pprint._format_funcs gets populated
        # 2) astropy.table.pprint._format_funcs gets used

        t = Table([[1.0, 2.0, 3.0], [3, 4, 5]], names=("a", "b"), masked=True)
        t["a"].mask = [True, False, True]
        # mathematical function
        t["a"].format = lambda x: str(x * 3.0)
        assert str(t["a"]) == " a \n---\n --\n6.0\n --"
        assert str(t["a"]) == " a \n---\n --\n6.0\n --"

    def test_column_format_func_with_special_masked(self):
        # run most of functions twice
        # 1) astropy.table.pprint._format_funcs gets populated
        # 2) astropy.table.pprint._format_funcs gets used

        t = Table([[1.0, 2.0, 3.0], [3, 4, 5]], names=("a", "b"), masked=True)
        t["a"].mask = [True, False, True]
        # mathematical function

        def format_func(x):
            if x is np.ma.masked:
                return "!!"
            else:
                return str(x * 3.0)

        t["a"].format = format_func
        assert str(t["a"]) == " a \n---\n !!\n6.0\n !!"
        assert str(t["a"]) == " a \n---\n !!\n6.0\n !!"

    def test_column_format_callable(self):
        # run most of functions twice
        # 1) astropy.table.pprint._format_funcs gets populated
        # 2) astropy.table.pprint._format_funcs gets used

        t = Table([[1.0, 2.0, 3.0], [3, 4, 5]], names=("a", "b"), masked=True)
        t["a"].mask = [True, False, True]

        # mathematical function
        class format:
            def __call__(self, x):
                return str(x * 3.0)

        t["a"].format = format()
        assert str(t["a"]) == " a \n---\n --\n6.0\n --"
        assert str(t["a"]) == " a \n---\n --\n6.0\n --"

    def test_column_format_func_wrong_number_args(self):
        t = Table([[1.0, 2.0], [3, 4]], names=("a", "b"), masked=True)
        t["a"].mask = [True, False]

        # function that expects wrong number of arguments
        def func(a, b):
            pass

        with pytest.raises(ValueError):
            t["a"].format = func

        # but if all are masked, it never gets called
        t["a"].mask = [True, True]
        assert str(t["a"]) == " a \n---\n --\n --"

    def test_column_format_func_multiD(self):
        arr = [np.array([[1, 2], [10, 20]], dtype="i8")]
        t = Table(arr, names=["a"], masked=True)
        t["a"].mask[0, 1] = True
        t["a"].mask[1, 1] = True
        # mathematical function
        t["a"].format = lambda x: str(x * 3.0)
        outstr = [
            "    a     ",
            "----------",
            " 3.0 .. --",
            "30.0 .. --",
        ]
        assert str(t["a"]).splitlines() == outstr
        assert str(t["a"]).splitlines() == outstr


def test_pprint_npfloat32():
    """
    Test for #148, that np.float32 cannot by itself be formatted as float,
    but has to be converted to a python float.
    """
    dat = np.array([1.0, 2.0], dtype=np.float32)
    t = Table([dat], names=["a"])
    t["a"].format = "5.2f"
    assert str(t["a"]) == "  a  \n-----\n 1.00\n 2.00"


def test_pprint_py3_bytes():
    """
    Test for #1346 and #4944. Make sure a bytestring (dtype=S<N>) in Python 3
    is printed correctly (without the "b" prefix like b'string').
    """
    val = bytes("val", encoding="utf-8")
    blah = "bläh".encode()
    dat = np.array([val, blah], dtype=[("col", "S10")])
    t = table.Table(dat)
    assert t["col"].pformat() == ["col ", "----", " val", "bläh"]


def test_pprint_structured():
    su = table.Column(
        [
            (1, (1.5, [1.6, 1.7])),
            (2, (2.5, [2.6, 2.7])),
        ],
        name="su",
        dtype=[
            ("i", np.int64),
            ("f", [("p0", np.float64), ("p1", np.float64, (2,))]),
        ],
    )
    assert su.pformat() == [
        "  su [i, f[p0, p1]]   ",
        "----------------------",
        "(1, (1.5, [1.6, 1.7]))",
        "(2, (2.5, [2.6, 2.7]))",
    ]
    t = table.Table([su])
    assert t.pformat() == su.pformat()
    assert repr(t).splitlines() == [
        "<Table length=2>",
        "      su [i, f[p0, p1]]       ",
        "(int64, (float64, float64[2]))",
        "------------------------------",
        "        (1, (1.5, [1.6, 1.7]))",
        "        (2, (2.5, [2.6, 2.7]))",
    ]


def test_pprint_structured_with_format():
    dtype = np.dtype([("par", "f8"), ("min", "f8"), ("id", "i4"), ("name", "U4")])
    c = table.Column(
        [
            (1.2345678, -20, 3, "bar"),
            (12.345678, 4.5678, 33, "foo"),
        ],
        dtype=dtype,
    )
    t = table.Table()
    t["a"] = [1, 2]
    t["c"] = c
    t["c"].info.format = "{par:6.2f} {min:5.1f} {id:03d} {name:4s}"
    exp = [
        " a  c [par, min, id, name]",
        "--- ----------------------",
        "  1    1.23 -20.0 003 bar ",
        "  2   12.35   4.6 033 foo ",
    ]
    assert t.pformat() == exp


def test_pprint_nameless_col():
    """Regression test for #2213, making sure a nameless column can be printed
    using None as the name.
    """
    col = table.Column([1.0, 2.0])
    assert str(col).startswith("None")


def test_html():
    """Test HTML printing"""
    dat = np.array([1.0, 2.0], dtype=np.float32)
    t = Table([dat], names=["a"])

    lines = t.pformat(html=True)
    assert lines == [
        f'<table id="table{id(t)}">',
        "<thead><tr><th>a</th></tr></thead>",
        "<tr><td>1.0</td></tr>",
        "<tr><td>2.0</td></tr>",
        "</table>",
    ]

    lines = t.pformat(html=True, tableclass="table-striped")
    assert lines == [
        f'<table id="table{id(t)}" class="table-striped">',
        "<thead><tr><th>a</th></tr></thead>",
        "<tr><td>1.0</td></tr>",
        "<tr><td>2.0</td></tr>",
        "</table>",
    ]

    lines = t.pformat(html=True, tableclass=["table", "table-striped"])
    assert lines == [
        f'<table id="table{id(t)}" class="table table-striped">',
        "<thead><tr><th>a</th></tr></thead>",
        "<tr><td>1.0</td></tr>",
        "<tr><td>2.0</td></tr>",
        "</table>",
    ]


def test_align():
    t = simple_table(2, kinds="iS")
    assert t.pformat() == [
        " a   b ",
        "--- ---",
        "  1   b",
        "  2   c",
    ]
    # Use column format attribute
    t["a"].format = "<"
    assert t.pformat() == [
        " a   b ",
        "--- ---",
        "1     b",
        "2     c",
    ]

    # Now override column format attribute with various combinations of align
    tpf = [" a   b ", "--- ---", " 1   b ", " 2   c "]
    for align in ("^", ["^", "^"], ("^", "^")):
        assert tpf == t.pformat(align=align)

    assert t.pformat(align="<") == [
        " a   b ",
        "--- ---",
        "1   b  ",
        "2   c  ",
    ]
    assert t.pformat(align="0=") == [
        " a   b ",
        "--- ---",
        "001 00b",
        "002 00c",
    ]

    assert t.pformat(align=["<", "^"]) == [
        " a   b ",
        "--- ---",
        "1    b ",
        "2    c ",
    ]

    # Now use fill characters.  Stress the system using a fill
    # character that is the same as an align character.
    t = simple_table(2, kinds="iS")

    assert t.pformat(align="^^") == [
        " a   b ",
        "--- ---",
        "^1^ ^b^",
        "^2^ ^c^",
    ]

    assert t.pformat(align="^>") == [
        " a   b ",
        "--- ---",
        "^^1 ^^b",
        "^^2 ^^c",
    ]

    assert t.pformat(align="^<") == [
        " a   b ",
        "--- ---",
        "1^^ b^^",
        "2^^ c^^",
    ]

    # Complicated interaction (same as narrative docs example)
    t1 = Table([[1.0, 2.0], [1, 2]], names=["column1", "column2"])
    t1["column1"].format = "#^.2f"

    assert t1.pformat() == [
        "column1 column2",
        "------- -------",
        "##1.00#       1",
        "##2.00#       2",
    ]

    assert t1.pformat(align="!<") == [
        "column1 column2",
        "------- -------",
        "1.00!!! 1!!!!!!",
        "2.00!!! 2!!!!!!",
    ]

    assert t1.pformat(align=[None, "!<"]) == [
        "column1 column2",
        "------- -------",
        "##1.00# 1!!!!!!",
        "##2.00# 2!!!!!!",
    ]

    # Zero fill
    t["a"].format = "+d"
    assert t.pformat(align="0=") == [
        " a   b ",
        "--- ---",
        "+01 00b",
        "+02 00c",
    ]

    with pytest.raises(ValueError):
        t.pformat(align=["fail"])

    with pytest.raises(TypeError):
        t.pformat(align=0)

    with pytest.raises(TypeError):
        t.pprint(align=0)

    # Make sure pprint() does not raise an exception
    t.pprint()

    with pytest.raises(ValueError):
        t.pprint(align=["<", "<", "<"])

    with pytest.raises(ValueError):
        t.pprint(align="x=")


def test_auto_format_func():
    """Test for #5802 (fix for #5800 where format_func key is not unique)"""
    t = Table([[1, 2] * u.m])
    t["col0"].format = "%f"
    t.pformat()  # Force caching of format function

    qt = QTable(t)
    qt.pformat()  # Generates exception prior to #5802


def test_decode_replace():
    """
    Test printing a bytestring column with a value that fails
    decoding to utf-8 and gets replaced by U+FFFD.  See
    https://docs.python.org/3/library/codecs.html#codecs.replace_errors
    """
    t = Table([[b"Z\xf0"]])
    assert t.pformat() == [
        "col0",
        "----",
        "  Z\ufffd",
    ]


class TestColumnsShowHide:
    """Tests of show and hide table columns"""

    def setup_method(self):
        self.t = simple_table(size=1, cols=4, kinds="i")

    @pytest.mark.parametrize("attr", ("pprint_exclude_names", "pprint_include_names"))
    def test_basic(self, attr):
        t = self.t
        assert (
            repr(getattr(Table, attr))
            == f"<PprintIncludeExclude name={attr} default=None>"
        )

        t_show_hide = getattr(t, attr)
        assert repr(t_show_hide) == f"<PprintIncludeExclude name={attr} value=None>"

        # Default value is None
        assert t_show_hide() is None

    def test_slice(self):
        t = self.t
        t.pprint_include_names = "a"
        t.pprint_exclude_names = "b"
        t2 = t[0:1]
        assert t2.pprint_include_names() == ("a",)
        assert t2.pprint_exclude_names() == ("b",)

    def test_copy(self):
        t = self.t
        t.pprint_include_names = "a"
        t.pprint_exclude_names = "b"

        t2 = t.copy()
        assert t2.pprint_include_names() == ("a",)
        assert t2.pprint_exclude_names() == ("b",)

        t2.pprint_include_names = "c"
        t2.pprint_exclude_names = "d"
        assert t.pprint_include_names() == ("a",)
        assert t.pprint_exclude_names() == ("b",)
        assert t2.pprint_include_names() == ("c",)
        assert t2.pprint_exclude_names() == ("d",)

    @pytest.mark.parametrize("attr", ("pprint_exclude_names", "pprint_include_names"))
    @pytest.mark.parametrize("value", ("z", ["a", "z"]))
    def test_setting(self, attr, value):
        t = self.t
        t_show_hide = getattr(t, attr)

        # Expected attribute value ('z',) or ('a', 'z')
        exp = (value,) if isinstance(value, str) else tuple(value)

        # Context manager, can include column names that do not exist
        with t_show_hide.set(value):
            assert t_show_hide() == exp
            assert t.meta["__attributes__"] == {attr: exp}
        assert t_show_hide() is None

        # Setting back to None clears out meta
        assert t.meta == {}

        # Do `t.pprint_include_names/hide = value`
        setattr(t, attr, value)
        assert t_show_hide() == exp

        # Clear attribute
        t_show_hide.set(None)
        assert t_show_hide() is None

        # Now use set() method
        t_show_hide.set(value)
        assert t_show_hide() == exp

        with t_show_hide.set(None):
            assert t_show_hide() is None
            assert t.meta == {}
        assert t_show_hide() == exp

    @pytest.mark.parametrize("attr", ("pprint_exclude_names", "pprint_include_names"))
    @pytest.mark.parametrize("value", ("z", ["a", "z"], ("a", "z")))
    def test_add_remove(self, attr, value):
        t = self.t
        t_show_hide = getattr(t, attr)

        # Expected attribute value ('z') or ('a', 'z')
        exp = (value,) if isinstance(value, str) else tuple(value)

        # add() method for str or list of str
        t_show_hide.add(value)
        assert t_show_hide() == exp

        # Adding twice has no effect
        t_show_hide.add(value)
        assert t_show_hide() == exp

        # Remove values (str or list of str). Reverts to None if all names are
        # removed.
        t_show_hide.remove(value)
        assert t_show_hide() is None

        # Remove just one name, possibly leaving a name.
        t_show_hide.add(value)
        t_show_hide.remove("z")
        assert t_show_hide() == (None if value == "z" else ("a",))

        # Cannot remove name not in the list
        t_show_hide.set(["a", "z"])
        with pytest.raises(ValueError, match=f"x not in {attr}"):
            t_show_hide.remove(("x", "z"))

    @pytest.mark.parametrize("attr", ("pprint_exclude_names", "pprint_include_names"))
    def test_rename(self, attr):
        t = self.t
        t_hide_show = getattr(t, attr)
        t_hide_show.set(["a", "b"])
        t.rename_column("a", "aa")
        assert t_hide_show() == ("aa", "b")

    @pytest.mark.parametrize("attr", ("pprint_exclude_names", "pprint_include_names"))
    def test_remove(self, attr):
        t = self.t
        t_hide_show = getattr(t, attr)
        t_hide_show.set(["a", "b"])
        del t["a"]
        assert t_hide_show() == ("b",)

    def test_serialization(self):
        # Serialization works for ECSV. Currently fails for FITS, works with
        # HDF5.
        t = self.t
        t.pprint_exclude_names = ["a", "y"]
        t.pprint_include_names = ["b", "z"]

        out = StringIO()
        ascii.write(t, out, format="ecsv")
        t2 = ascii.read(out.getvalue(), format="ecsv")

        assert t2.pprint_exclude_names() == ("a", "y")
        assert t2.pprint_include_names() == ("b", "z")

    def test_output(self):
        """Test that pprint_include/exclude_names actually changes the print output"""
        t = self.t
        exp = [
            " b   d ",
            "--- ---",
            "  2   4",
        ]

        with t.pprint_exclude_names.set(["a", "c"]):
            out = t.pformat()
        assert out == exp

        with t.pprint_include_names.set(["b", "d"]):
            out = t.pformat()
        assert out == exp

        with t.pprint_exclude_names.set(["a", "c"]):
            out = t.pformat()
        assert out == exp

        with t.pprint_include_names.set(["b", "d"]):
            out = t.pformat()
        assert out == exp

        with (
            t.pprint_include_names.set(["b", "c", "d"]),
            t.pprint_exclude_names.set(["c"]),
        ):
            out = t.pformat()
        assert out == exp

    def test_output_globs(self):
        """Test that pprint_include/exclude_names works with globs (fnmatch)"""
        t = self.t
        t["a2"] = 1
        t["a23"] = 2

        # Show only the a* columns
        exp = [
            " a   a2 a23",
            "--- --- ---",
            "  1   1   2",
        ]
        with t.pprint_include_names.set("a*"):
            out = t.pformat()
        assert out == exp

        # Show a* but exclude a??
        exp = [
            " a   a2",
            "--- ---",
            "  1   1",
        ]
        with t.pprint_include_names.set("a*"), t.pprint_exclude_names.set("a??"):
            out = t.pformat()
        assert out == exp

        # Exclude a??
        exp = [
            " a   b   c   d   a2",
            "--- --- --- --- ---",
            "  1   2   3   4   1",
        ]
        with t.pprint_exclude_names.set("a??"):
            out = t.pformat()
        assert out == exp


def test_embedded_newline_tab():
    """Newlines and tabs are escaped in table repr"""
    t = Table(
        rows=[
            ["a", "b \n c \t \n d"],
            ["x", "y\n"],
        ]
    )
    exp = [
        r"col0      col1     ",
        r"---- --------------",
        r"   a b \n c \t \n d",
        r"   x            y\n",
    ]
    assert t.pformat() == exp


def test_multidims_with_zero_dim():
    """Test of fix for #13836 when a zero-dim column is present"""
    t = Table()
    t["a"] = ["a", "b"]
    t["b"] = np.ones(shape=(2, 0, 1), dtype=np.float64)
    exp = [
        " a        b      ",
        "str1 float64[0,1]",
        "---- ------------",
        "   a             ",
        "   b             ",
    ]
    assert t.pformat(show_dtype=True) == exp


def test_zero_length_string():
    data = np.array([("", 12)], dtype=[("a", "S"), ("b", "i4")])
    t = Table(data, copy=False)
    exp = [
        "  a      b  ",
        "bytes0 int32",
        "------ -----",
        "          12",
    ]
    assert t.pformat(show_dtype=True) == exp



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import io
import os
import re
import subprocess
import sys
from contextlib import nullcontext
from inspect import cleandoc
from pathlib import Path
from threading import Lock

import pytest

from astropy.config import configuration, create_config_file, paths, set_temp_config
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyDeprecationWarning, AstropyUserWarning

OLD_CONFIG = {}

_IGNORE_CONFIG_PATHS_GLOBAL_STATE_LOCK = Lock()


@pytest.fixture
def ignore_config_paths_global_state(monkeypatch, tmp_path_factory):
    # ignore global state of the test session
    # and preserve thread safety across all users of this fixture
    with _IGNORE_CONFIG_PATHS_GLOBAL_STATE_LOCK:
        monkeypatch.delenv("XDG_CACHE_HOME", raising=False)
        monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)

        monkeypatch.setattr(paths.set_temp_cache, "_temp_path", None)
        monkeypatch.setattr(paths.set_temp_config, "_temp_path", None)

        # also mock $HOME as it's part of the global state taken into account
        # for path detection
        mock_home_dir = tmp_path_factory.mktemp("MOCK_HOME")

        def mock_home():
            return mock_home_dir

        monkeypatch.setattr(Path, "home", mock_home)

        yield


def setup_module():
    OLD_CONFIG.clear()
    OLD_CONFIG.update(configuration._cfgobjs)


def teardown_module():
    configuration._cfgobjs.clear()
    configuration._cfgobjs.update(OLD_CONFIG)


@pytest.mark.usefixtures("ignore_config_paths_global_state")
def test_paths():
    assert "astropy" in paths.get_config_dir()
    assert "astropy" in paths.get_cache_dir()
    assert str(paths.get_config_dir_path()) == paths.get_config_dir()
    assert str(paths.get_cache_dir_path()) == paths.get_cache_dir()
    assert paths.get_config_dir_path().is_absolute()
    assert paths.get_cache_dir_path().is_absolute()

    assert "testpkg" in paths.get_config_dir(rootname="testpkg")
    assert "testpkg" in paths.get_cache_dir(rootname="testpkg")


ENV_VAR_AND_FUNC = [
    pytest.param(
        "XDG_CACHE_HOME",
        paths.get_cache_dir_path,
        id="xdg-cache",
    ),
    pytest.param(
        "XDG_CONFIG_HOME",
        paths.get_config_dir_path,
        id="xdg-config",
    ),
]


@pytest.mark.parametrize("env_var, func", ENV_VAR_AND_FUNC)
@pytest.mark.usefixtures("ignore_config_paths_global_state")
def test_xdg_variables(monkeypatch, tmp_path, env_var, func):
    # Regression test for #17514 - XDG_CACHE_HOME had no effect
    target_dir = tmp_path / "astropy"
    target_dir.mkdir()
    monkeypatch.setenv(env_var, str(tmp_path))
    assert func() == target_dir


@pytest.mark.parametrize("env_var, func", ENV_VAR_AND_FUNC)
@pytest.mark.usefixtures("ignore_config_paths_global_state")
def test_env_variables_missing_dir(monkeypatch, tmp_path, env_var, func):
    default_path = func()
    target_dir = tmp_path / "nonexistent"
    monkeypatch.setenv(env_var, str(target_dir))

    expected_msg = (
        rf"^{env_var} is set to '{re.escape(str(target_dir))}', "
        rf"but no such directory was found\. "
        r"This environment variable will be ignored\.$"
    )
    with pytest.warns(AstropyUserWarning, match=expected_msg):
        new_path = func()

    assert new_path == default_path


def get_directory_type(func) -> str:
    for t in ["cache", "config"]:
        if t in func.__name__:
            return t
    raise RuntimeError


@pytest.mark.parametrize("env_var, func", ENV_VAR_AND_FUNC)
@pytest.mark.usefixtures("ignore_config_paths_global_state")
def test_env_variables_missing_subdir_and_default(monkeypatch, tmp_path, env_var, func):
    # have the env var point to a writable location,
    # where an 'astropy' subdir is missing, but can be created silently
    default_parent_dir = Path.home() / ".astropy"
    assert not default_parent_dir.exists()

    target_dir = tmp_path
    monkeypatch.setenv(env_var, str(target_dir))

    expected_location = target_dir / "astropy"
    default_location = default_parent_dir / get_directory_type(func)

    assert not expected_location.exists()

    if sys.platform.startswith("win"):
        # FIXME: legacy behavior is to never symlink on windows...
        # should we never symlink anywhere instead for consistency
        # (not to mention simplicity) ?
        expected_msg = (
            rf"^{env_var} is set to '{re.escape(str(target_dir))}', "
            rf"but is not supported on Windows\. "
            r"This environment variable will be ignored\.$"
        )
        ctx = pytest.warns(AstropyUserWarning, match=expected_msg)
    else:
        ctx = nullcontext()

    with ctx:
        ret = func()

    assert ret.is_dir()

    # FIXME: should the environment variable *completely* take precedence here,
    # and should we *avoid* silently creating (and linking) the default location ?
    # see https://github.com/astropy/astropy/issues/18791
    assert ret == default_location

    if sys.platform.startswith("win"):
        assert not expected_location.exists()
        return

    assert expected_location.is_dir()
    assert default_location.is_dir()
    assert expected_location.exists()
    assert expected_location.is_dir()
    assert expected_location.is_symlink()
    assert expected_location.resolve() == default_location


@pytest.mark.parametrize("env_var, func", ENV_VAR_AND_FUNC)
@pytest.mark.usefixtures("ignore_config_paths_global_state")
def test_env_variables_missing_subdir_but_default_exists(
    monkeypatch, tmp_path, env_var, func
):
    # have the env var point to a writable location,
    # where an 'astropy' subdir is missing, but can be created silently
    default_parent_dir = Path.home() / ".astropy"

    target_dir = tmp_path
    monkeypatch.setenv(env_var, str(target_dir))

    expected_location = target_dir / "astropy"
    default_location = default_parent_dir / get_directory_type(func)

    # this is the crucial variation in setup
    # as compared to test_env_variables_missing_subdir_and_default
    default_location.mkdir(parents=True)

    assert not expected_location.exists()

    expected_msg_template = (
        rf"^{env_var} is set to '{re.escape(str(target_dir))}', "
        r"but {reason}. "
        r"This environment variable will be ignored\.$"
    )
    if sys.platform.startswith("win"):
        # FIXME: legacy behavior is to never symlink on windows...
        # should we never symlink anywhere instead for consistency
        # (not to mention simplicity) ?
        reason = "is not supported on Windows"
    else:
        reason = (
            rf"the default location, {re.escape(str(default_location))}, "
            "already exists, and takes precedence"
        )

    expected_msg = expected_msg_template.format(reason=reason)

    with pytest.warns(AstropyUserWarning, match=expected_msg):
        ret = func()

    assert ret.is_dir()

    # FIXME: should the environment variable *completely* take precedent here,
    # ignoring the existence of default_parent_dir ?
    # see https://github.com/astropy/astropy/issues/18791
    assert not expected_location.exists()


@pytest.mark.parametrize("env_var, func", ENV_VAR_AND_FUNC)
@pytest.mark.usefixtures("ignore_config_paths_global_state")
def test_env_variables_setup_file_exists(monkeypatch, tmp_path, env_var, func):
    # check what happens if we request a location that's already
    # taken, but is a file
    default_path = func()
    target_dir = tmp_path / "subdir"
    target_dir.mkdir()
    expected_path = target_dir / "astropy"
    expected_path.touch()  # create a file

    monkeypatch.setenv(env_var, str(target_dir))

    expected_msg = (
        rf"^{env_var} is set to '{re.escape(str(target_dir))}', "
        rf"but this directory already contains a file under '{expected_path.name}', "
        r"where a directory was expected\. "
        r"This environment variable will be ignored\.$"
    )
    with pytest.warns(AstropyUserWarning, match=expected_msg):
        new_path = func()
    assert new_path == default_path


@pytest.mark.usefixtures("ignore_config_paths_global_state")
def test_set_temp_config(tmp_path):
    # Check that we start in an understood state.
    assert configuration._cfgobjs == OLD_CONFIG

    orig_config_dir = paths.get_config_dir(rootname="astropy")
    (temp_config_dir := tmp_path / "config").mkdir()
    temp_astropy_config = temp_config_dir / "astropy"

    # Test decorator mode
    @paths.set_temp_config(temp_config_dir)
    def test_func():
        assert paths.get_config_dir(rootname="astropy") == str(temp_astropy_config)
        assert paths.get_config_dir_path(rootname="astropy") == temp_astropy_config

        # Test temporary restoration of original default
        with paths.set_temp_config() as d:
            assert d == orig_config_dir == paths.get_config_dir(rootname="astropy")

    test_func()

    # Test context manager mode (with cleanup)
    with paths.set_temp_config(temp_config_dir, delete=True):
        assert paths.get_config_dir(rootname="astropy") == str(temp_astropy_config)
        assert paths.get_config_dir_path(rootname="astropy") == temp_astropy_config

    assert not temp_config_dir.exists()
    # Check that we have returned to our old configuration.
    assert configuration._cfgobjs == OLD_CONFIG


@pytest.mark.usefixtures("ignore_config_paths_global_state")
def test_set_temp_cache(tmp_path):
    orig_cache_dir = paths.get_cache_dir(rootname="astropy")
    (temp_cache_dir := tmp_path / "cache").mkdir()
    temp_astropy_cache = temp_cache_dir / "astropy"

    # Test decorator mode
    @paths.set_temp_cache(temp_cache_dir)
    def test_func():
        assert paths.get_cache_dir(rootname="astropy") == str(temp_astropy_cache)

        # Test temporary restoration of original default
        with paths.set_temp_cache() as d:
            assert d == orig_cache_dir == paths.get_cache_dir(rootname="astropy")

    test_func()

    # Test context manager mode (with cleanup)
    with paths.set_temp_cache(temp_cache_dir, delete=True):
        assert paths.get_cache_dir(rootname="astropy") == str(temp_astropy_cache)

    assert not temp_cache_dir.exists()


def test_set_temp_cache_resets_on_exception(tmp_path):
    """Test for regression of  bug #9704"""
    t = paths.get_cache_dir()
    (a := tmp_path / "a").write_text("not a good cache\n")
    with pytest.raises(OSError), paths.set_temp_cache(a):
        pass
    assert t == paths.get_cache_dir()


@pytest.mark.usefixtures("ignore_config_paths_global_state")
def test_config_file():
    from astropy.config.configuration import get_config, reload_config

    apycfg = get_config("astropy")
    assert apycfg.filename.endswith("astropy.cfg")

    cfgsec = get_config("astropy.config")
    assert cfgsec.depth == 1
    assert cfgsec.name == "config"
    assert cfgsec.parent.filename.endswith("astropy.cfg")

    # try with a different package name, still inside astropy config dir:
    testcfg = get_config("testpkg", rootname="astropy")
    parts = os.path.normpath(testcfg.filename).split(os.sep)
    assert ".astropy" in parts or "astropy" in parts
    assert parts[-1] == "testpkg.cfg"
    configuration._cfgobjs["testpkg"] = None  # HACK

    # try with a different package name, no specified root name (should
    #   default to astropy):
    testcfg = get_config("testpkg")
    parts = os.path.normpath(testcfg.filename).split(os.sep)
    assert ".astropy" in parts or "astropy" in parts
    assert parts[-1] == "testpkg.cfg"
    configuration._cfgobjs["testpkg"] = None  # HACK

    # try with a different package name, specified root name:
    testcfg = get_config("testpkg", rootname="testpkg")
    parts = os.path.normpath(testcfg.filename).split(os.sep)
    assert ".testpkg" in parts or "testpkg" in parts
    assert parts[-1] == "testpkg.cfg"
    configuration._cfgobjs["testpkg"] = None  # HACK

    # try with a subpackage with specified root name:
    testcfg_sec = get_config("testpkg.somemodule", rootname="testpkg")
    parts = os.path.normpath(testcfg_sec.parent.filename).split(os.sep)
    assert ".testpkg" in parts or "testpkg" in parts
    assert parts[-1] == "testpkg.cfg"
    configuration._cfgobjs["testpkg"] = None  # HACK

    reload_config("astropy")


def check_config(conf):
    # test that the output contains some lines that we expect
    assert "# unicode_output = False" in conf
    assert "[io.fits]" in conf
    assert "[table]" in conf
    assert "# replace_warnings = ," in conf
    assert "[table.jsviewer]" in conf
    assert "# css_urls = https://cdn.datatables.net/2.1.8/css/dataTables.dataTables.min.css" in conf  # fmt: skip
    assert "[visualization.wcsaxes]" in conf
    assert "## Whether to log exceptions before raising them." in conf
    assert "# log_exceptions = False" in conf


def test_generate_config(tmp_path):
    from astropy.config.configuration import generate_config

    out = io.StringIO()
    generate_config("astropy", out)
    conf = out.getvalue()

    outfile = tmp_path / "astropy.cfg"
    generate_config("astropy", outfile)
    with open(outfile) as fp:
        conf2 = fp.read()

    for c in (conf, conf2):
        check_config(c)


def test_generate_config2(tmp_path):
    """Test that generate_config works with the default filename."""

    with set_temp_config(tmp_path):
        from astropy.config.configuration import generate_config

        generate_config("astropy")

    assert os.path.exists(tmp_path / "astropy" / "astropy.cfg")

    with open(tmp_path / "astropy" / "astropy.cfg") as fp:
        conf = fp.read()

    check_config(conf)


def test_generate_config_subclasses(tmp_path):
    """Test that generate_config works with subclasses of ConfigNamespace."""
    from astropy.config.configuration import ConfigItem, ConfigNamespace

    class MyPackageNamespace(ConfigNamespace):
        pass

    class RecursiveTestConf(MyPackageNamespace):
        ti = ConfigItem(5, "this is a Description")

    with set_temp_config(tmp_path):
        from astropy.config.configuration import generate_config

        generate_config("astropy")

    assert os.path.exists(tmp_path / "astropy" / "astropy.cfg")

    with open(tmp_path / "astropy" / "astropy.cfg") as fp:
        conf = fp.read()

    assert "# ti = 5" in conf


def test_create_config_file(tmp_path, caplog):
    with set_temp_config(tmp_path):
        create_config_file("astropy")

    # check that the config file has been created
    assert (
        "The configuration file has been successfully written"
        in caplog.records[0].message
    )
    assert os.path.exists(tmp_path / "astropy" / "astropy.cfg")

    with open(tmp_path / "astropy" / "astropy.cfg") as fp:
        conf = fp.read()
    check_config(conf)

    caplog.clear()

    # now modify the config file
    conf = conf.replace("# unicode_output = False", "unicode_output = True")
    with open(tmp_path / "astropy" / "astropy.cfg", mode="w") as fp:
        fp.write(conf)

    with set_temp_config(tmp_path):
        create_config_file("astropy")

    # check that the config file has not been overwritten since it was modified
    assert (
        "The configuration file already exists and seems to have been customized"
        in caplog.records[0].message
    )

    caplog.clear()

    with set_temp_config(tmp_path):
        create_config_file("astropy", overwrite=True)

    # check that the config file has been overwritten
    assert (
        "The configuration file has been successfully written"
        in caplog.records[0].message
    )


def test_configitem():
    from astropy.config.configuration import ConfigItem, ConfigNamespace, get_config

    ci = ConfigItem(34, "this is a Description")

    class Conf(ConfigNamespace):
        tstnm = ci

    conf = Conf()

    assert ci.module == "astropy.config.tests.test_configs"
    assert ci() == 34
    assert ci.description == "this is a Description"

    assert conf.tstnm == 34

    assert str(conf) == cleandoc(
        """
        Configuration parameters for `astropy.config.tests.test_configs`

        ConfigItem: tstnm
          cfgtype='integer'
          defaultvalue=34
          description='this is a Description'
          module=astropy.config.tests.test_configs
          value=34
        """
    )

    sec = get_config(ci.module)
    assert sec["tstnm"] == 34

    ci.description = "updated Descr"
    ci.set(32)
    assert ci() == 32

    # It's useful to go back to the default to allow other test functions to
    # call this one and still be in the default configuration.
    ci.description = "this is a Description"
    ci.set(34)
    assert ci() == 34

    # Test iterator for one-item namespace
    result = list(conf)
    assert result == ["tstnm"]
    result = list(conf.keys())
    assert result == ["tstnm"]
    result = list(conf.values())
    assert result == [ci]
    result = list(conf.items())
    assert result == [("tstnm", ci)]


def test_configitem_types():
    from astropy.config.configuration import ConfigItem, ConfigNamespace

    ci1 = ConfigItem(34)
    ci2 = ConfigItem(34.3)
    ci3 = ConfigItem(True)
    ci4 = ConfigItem("astring")

    class Conf(ConfigNamespace):
        tstnm1 = ci1
        tstnm2 = ci2
        tstnm3 = ci3
        tstnm4 = ci4

    conf = Conf()

    assert isinstance(conf.tstnm1, int)
    assert isinstance(conf.tstnm2, float)
    assert isinstance(conf.tstnm3, bool)
    assert isinstance(conf.tstnm4, str)

    with pytest.raises(TypeError):
        conf.tstnm1 = 34.3
    conf.tstnm2 = 12  # this would should succeed as up-casting
    with pytest.raises(TypeError):
        conf.tstnm3 = "fasd"
    with pytest.raises(TypeError):
        conf.tstnm4 = 546.245

    # Test iterator for multi-item namespace. Assume ordered by insertion order.
    item_names = list(conf)
    assert item_names == ["tstnm1", "tstnm2", "tstnm3", "tstnm4"]
    result = list(conf.keys())
    assert result == item_names
    result = list(conf.values())
    assert result == [ci1, ci2, ci3, ci4]
    result = list(conf.items())
    assert result == [
        ("tstnm1", ci1),
        ("tstnm2", ci2),
        ("tstnm3", ci3),
        ("tstnm4", ci4),
    ]


def test_configitem_options(tmp_path):
    from astropy.config.configuration import ConfigItem, ConfigNamespace, get_config

    cio = ConfigItem(["op1", "op2", "op3"])

    class Conf(ConfigNamespace):
        tstnmo = cio

    sec = get_config(cio.module)

    assert isinstance(cio(), str)
    assert cio() == "op1"
    assert sec["tstnmo"] == "op1"

    cio.set("op2")
    with pytest.raises(TypeError):
        cio.set("op5")
    assert sec["tstnmo"] == "op2"

    # now try saving
    apycfg = sec
    while apycfg.parent is not apycfg:
        apycfg = apycfg.parent
    f = tmp_path / "astropy.cfg"
    with open(f, "wb") as fd:
        apycfg.write(fd)

    assert "tstnmo = op2" in f.read_text().splitlines()


def test_help(capsys):
    from astropy import conf

    use_color_msg = cleandoc(
        """
        ConfigItem: use_color
          cfgtype='boolean'
          defaultvalue={is_not_windows}
          description='When True, use ANSI color escape sequences when writing to the console.'
          module=astropy
          value={is_not_windows}
        """
    ).format(is_not_windows=sys.platform != "win32")
    conf.help("use_color")
    assert capsys.readouterr().out == use_color_msg + "\n"

    conf.help()
    help_text = capsys.readouterr().out
    assert help_text.startswith(
        "Configuration parameters for `astropy`.\n\nConfigItem: unicode_output"
    )
    assert use_color_msg in help_text


def test_help_invalid_config_item():
    from astropy import conf

    with pytest.raises(
        KeyError,
        match=(
            "'bad_name' is not among configuration items "
            r"\('unicode_output', 'use_color', 'max_lines', 'max_width'\)"
        ),
    ):
        conf.help("bad_name")


@pytest.mark.usefixtures("ignore_config_paths_global_state")
def test_config_noastropy_fallback(monkeypatch):
    """
    Tests to make sure configuration items fall back to their defaults when
    there's a problem accessing the astropy directory
    """

    # make sure the _find_or_create_root_dir function fails as though the
    # astropy dir could not be accessed
    @classmethod
    def osraiser(cls, linkto, pkgname=None):
        raise OSError

    monkeypatch.setattr(paths._SetTempPath, "_find_or_create_root_dir", osraiser)

    # also have to make sure the stored configuration objects are cleared
    monkeypatch.setattr(configuration, "_cfgobjs", {})

    # make sure the config dir search fails
    with pytest.raises(OSError):
        paths.get_config_dir(rootname="astropy")

    with pytest.raises(OSError):
        paths.get_config_dir_path(rootname="astropy")

    # now run the basic tests, and make sure the warning about no astropy
    # is present
    test_configitem()


def test_configitem_setters():
    from astropy.config.configuration import ConfigItem, ConfigNamespace

    class Conf(ConfigNamespace):
        tstnm12 = ConfigItem(42, "this is another Description")

    conf = Conf()

    assert conf.tstnm12 == 42
    with conf.set_temp("tstnm12", 45):
        assert conf.tstnm12 == 45
    assert conf.tstnm12 == 42

    conf.tstnm12 = 43
    assert conf.tstnm12 == 43

    with conf.set_temp("tstnm12", 46):
        assert conf.tstnm12 == 46

    # Make sure it is reset even with Exception
    try:
        with conf.set_temp("tstnm12", 47):
            raise Exception
    except Exception:
        pass

    assert conf.tstnm12 == 43


def test_empty_config_file():
    from astropy.config.configuration import is_unedited_config_file

    def get_content(fn):
        with open(get_pkg_data_filename(fn), encoding="latin-1") as fd:
            return fd.read()

    content = get_content("data/empty.cfg")
    assert is_unedited_config_file(content)

    content = get_content("data/not_empty.cfg")
    assert not is_unedited_config_file(content)


class TestAliasRead:
    def setup_class(self):
        configuration._override_config_file = get_pkg_data_filename("data/alias.cfg")

    def test_alias_read(self):
        from astropy.utils.data import conf

        with pytest.warns(
            AstropyDeprecationWarning,
            match=r"Config parameter 'name_resolve_timeout' in section "
            r"\[coordinates.name_resolve\].*",
        ) as w:
            conf.reload()
            assert conf.remote_timeout == 42

        assert len(w) == 1

    def teardown_class(self):
        from astropy.utils.data import conf

        configuration._override_config_file = None
        conf.reload()


def test_configitem_unicode():
    from astropy.config.configuration import ConfigItem, ConfigNamespace, get_config

    cio = ConfigItem("ასტრონომიის")

    class Conf(ConfigNamespace):
        tstunicode = cio

    sec = get_config(cio.module)

    assert isinstance(cio(), str)
    assert cio() == "ასტრონომიის"
    assert sec["tstunicode"] == "ასტრონომიის"


def test_warning_move_to_top_level():
    # Check that the warning about deprecation config items in the
    # file works.  See #2514
    from astropy import conf

    configuration._override_config_file = get_pkg_data_filename("data/deprecated.cfg")

    try:
        with pytest.warns(AstropyDeprecationWarning) as w:
            conf.reload()
            conf.max_lines
        assert len(w) == 1
    finally:
        configuration._override_config_file = None
        conf.reload()


def test_no_home():
    # "import astropy" fails when neither $HOME or $XDG_CONFIG_HOME
    # are set.  To test, we unset those environment variables for a
    # subprocess and try to import astropy.

    test_path = os.path.dirname(__file__)
    astropy_path = os.path.abspath(os.path.join(test_path, "..", "..", ".."))

    env = os.environ.copy()
    paths = [astropy_path]
    if env.get("PYTHONPATH"):
        paths.append(env.get("PYTHONPATH"))
    env["PYTHONPATH"] = os.pathsep.join(paths)

    for val in ["HOME", "XDG_CONFIG_HOME"]:
        if val in env:
            del env[val]

    retcode = subprocess.check_call([sys.executable, "-c", "import astropy"], env=env)

    assert retcode == 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This package contains the coordinate frames implemented by astropy.

Users shouldn't use this module directly, but rather import from the
`astropy.coordinates` module.  While it is likely to exist for the long-term,
the existence of this package and details of its organization should be
considered an implementation detail, and is not guaranteed to hold for future
versions of astropy.

Notes
-----
The builtin frame classes are all imported automatically into this package's
namespace, so there's no need to access the sub-modules directly.

To implement a new frame in Astropy, a developer should add the frame as a new
module in this package.  Any "self" transformations (i.e., those that transform
from one frame to another frame of the same class) should be included in that
module.  Transformation functions connecting the new frame to other frames
should be in a separate module, which should be imported in this package's
``__init__.py`` to ensure the transformations are hooked up when this package is
imported.  Placing the transformation functions in separate modules avoids
circular dependencies, because they need references to the frame classes.
"""

from astropy.coordinates.baseframe import frame_transform_graph

from .altaz import AltAz
from .baseradec import BaseRADecFrame
from .cirs import CIRS
from .ecliptic import (
    BarycentricMeanEcliptic,
    BarycentricTrueEcliptic,
    BaseEclipticFrame,
    CustomBarycentricEcliptic,
    GeocentricMeanEcliptic,
    GeocentricTrueEcliptic,
    HeliocentricEclipticIAU76,
    HeliocentricMeanEcliptic,
    HeliocentricTrueEcliptic,
)
from .equatorial import TEME, TETE
from .fk4 import FK4, FK4NoETerms
from .fk5 import FK5
from .galactic import Galactic
from .galactocentric import Galactocentric, galactocentric_frame_defaults
from .gcrs import GCRS, PrecessedGeocentric
from .hadec import HADec
from .hcrs import HCRS
from .icrs import ICRS
from .itrs import ITRS
from .skyoffset import SkyOffsetFrame
from .supergalactic import Supergalactic

# isort: split
# need to import transformations so that they get registered in the graph
from . import (
    cirs_observed_transforms,
    fk4_fk5_transforms,
    galactic_transforms,
    icrs_cirs_transforms,
    icrs_fk5_transforms,
    icrs_observed_transforms,
    intermediate_rotation_transforms,
    itrs_observed_transforms,
    supergalactic_transforms,
)

# isort: split
from . import ecliptic_transforms

# isort: split
# Import this after importing other frames, since this requires various
# transformations to set up the LSR frames
from .lsr import LSR, LSRD, LSRK, GalacticLSR

# we define an __all__ because otherwise the transformation modules
# get included.  Note that the order here determines the order in the
# documentation of the built-in frames (see make_transform_graphs_docs).

# ignoring RUF022 here because case-sensitive  alphanumeric sorting doesn't work here
# (as of ruff 0.8). RUF100 also ignored so older versions of ruff don't remove
# the first noqa comment
# ref https://github.com/astropy/astropy/pull/17437#discussion_r1856780149
__all__ = [  # noqa: RUF022, RUF100
    "ICRS",
    "FK5",
    "FK4",
    "FK4NoETerms",
    "Galactic",
    "Galactocentric",
    "Supergalactic",
    "AltAz",
    "HADec",
    "GCRS",
    "CIRS",
    "ITRS",
    "HCRS",
    "TEME",
    "TETE",
    "PrecessedGeocentric",
    "GeocentricMeanEcliptic",
    "BarycentricMeanEcliptic",
    "HeliocentricMeanEcliptic",
    "GeocentricTrueEcliptic",
    "BarycentricTrueEcliptic",
    "HeliocentricTrueEcliptic",
    "HeliocentricEclipticIAU76",
    "CustomBarycentricEcliptic",
    "LSR",
    "LSRK",
    "LSRD",
    "GalacticLSR",
    "SkyOffsetFrame",
    "BaseEclipticFrame",
    "BaseRADecFrame",
    "galactocentric_frame_defaults",
    "make_transform_graph_docs",
]


def _get_doc_header(cls):
    """Get the first line of a docstring.

    Skips possible empty first lines, and then combine following text until
    the first period or a fully empty line.
    """
    out = []

    # NOTE: cls.__doc__ is None for -OO flag
    if not cls.__doc__:
        return ""

    for line in cls.__doc__.splitlines():
        if line:
            parts = line.split(".")
            out.append(parts[0].strip())
            if len(parts) > 1:
                break
        elif out:
            break
    return " ".join(out) + "."


def make_transform_graph_docs(transform_graph):
    """
    Generates a string that can be used in other docstrings to include a
    transformation graph, showing the available transforms and
    coordinate systems.

    Parameters
    ----------
    transform_graph : `~astropy.coordinates.TransformGraph`

    Returns
    -------
    docstring : str
        A string that can be added to the end of a docstring to show the
        transform graph.
    """
    from textwrap import dedent

    coosys = {
        (cls := transform_graph.lookup_name(item)).__name__: cls
        for item in transform_graph.get_names()
    }

    # currently, all of the priorities are set to 1, so we don't need to show
    #   then in the transform graph.
    graphstr = transform_graph.to_dot_graph(
        addnodes=list(coosys.values()), priorities=False
    )

    docstr = """
    The diagram below shows all of the built in coordinate systems,
    their aliases (useful for converting other coordinates to them using
    attribute-style access) and the pre-defined transformations between
    them.  The user is free to override any of these transformations by
    defining new transformations between these systems, but the
    pre-defined transformations should be sufficient for typical usage.

    The color of an edge in the graph (i.e., the transformations between two
    frames) is set by the type of transformation; the legend box defines the
    mapping from transform class name to color.

    .. Wrap the graph in a div with a custom class to allow theming.
    .. container:: frametransformgraph

        .. graphviz::

    """

    docstr = dedent(docstr) + "        " + graphstr.replace("\n", "\n        ")

    # colors are in dictionary mapping transform class to color
    from astropy.coordinates.transformations.graph import trans_to_color

    html_list_items = []
    for cls, color in trans_to_color.items():
        block = f"""
            <li style='list-style: none;'>
                <p style="font-size: 12px;line-height: 24px;font-weight: normal;color: #848484;padding: 0;margin: 0;">
                    <b>{cls.__name__}:</b>
                    <span style="font-size: 24px; color: {color};"><b>➝</b></span>
                </p>
            </li>
        """
        html_list_items.append(block)

    nl = "\n"
    graph_legend = f"""
    .. raw:: html

        <ul class="cooframelegend">
            {nl.join(html_list_items)}
        </ul>
    """
    docstr = docstr + dedent(graph_legend)

    # Add table with built-in frame classes.
    template = """
       * - `~astropy.coordinates.{}`
         - {}
    """
    table = """
    Built-in Frame Classes
    ^^^^^^^^^^^^^^^^^^^^^^
    .. list-table::
       :widths: 20 80
    """ + "".join(
        template.format(name, _get_doc_header(coosys[name]))
        for name in __all__
        if name in coosys
    )

    return docstr + dedent(table)


_transform_graph_docs = make_transform_graph_docs(frame_transform_graph)

# Here, we override the module docstring so that sphinx renders the transform
# graph without the developer documentation in the main docstring above.
__doc__ = _transform_graph_docs



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
``fitsheader`` is a command line script based on astropy.io.fits for printing
the header(s) of one or more FITS file(s) to the standard output in a human-
readable format.

Example uses of fitsheader:

1. Print the header of all the HDUs of a .fits file::

    $ fitsheader filename.fits

2. Print the header of the third and fifth HDU extension::

    $ fitsheader --extension 3 --extension 5 filename.fits

3. Print the header of a named extension, e.g. select the HDU containing
   keywords EXTNAME='SCI' and EXTVER='2'::

    $ fitsheader --extension "SCI,2" filename.fits

4. Print only specific keywords::

    $ fitsheader --keyword BITPIX --keyword NAXIS filename.fits

5. Print keywords NAXIS, NAXIS1, NAXIS2, etc using a wildcard::

    $ fitsheader --keyword NAXIS* filename.fits

6. Dump the header keywords of all the files in the current directory into a
   machine-readable csv file::

    $ fitsheader --table ascii.csv *.fits > keywords.csv

7. Specify hierarchical keywords with the dotted or spaced notation::

    $ fitsheader --keyword ESO.INS.ID filename.fits
    $ fitsheader --keyword "ESO INS ID" filename.fits

8. Compare the headers of different fits files, following ESO's ``fitsort``
   format::

    $ fitsheader --fitsort --extension 0 --keyword ESO.INS.ID *.fits

9. Same as above, sorting the output along a specified keyword::

    $ fitsheader -f -s DATE-OBS -e 0 -k DATE-OBS -k ESO.INS.ID *.fits

10. Sort first by OBJECT, then DATE-OBS::

    $ fitsheader -f -s OBJECT -s DATE-OBS *.fits

Note that compressed images (HDUs of type
:class:`~astropy.io.fits.CompImageHDU`) really have two headers: a real
BINTABLE header to describe the compressed data, and a fake IMAGE header
representing the image that was compressed. Astropy returns the latter by
default. You must supply the ``--compressed`` option if you require the real
header that describes the compression.

With Astropy installed, please run ``fitsheader --help`` to see the full usage
documentation.
"""

import argparse
import sys

import numpy as np

from astropy import __version__, log
from astropy.io import fits
from astropy.io.fits import CompImageHDU

DESCRIPTION = """
Print the header(s) of a FITS file. Optional arguments allow the desired
extension(s), keyword(s), and output format to be specified.
Note that in the case of a compressed image, the decompressed header is
shown by default.

This script is part of the Astropy package. See
https://docs.astropy.org/en/latest/io/fits/usage/scripts.html#module-astropy.io.fits.scripts.fitsheader
for further documentation.
""".strip()


class ExtensionNotFoundException(Exception):
    """Raised if an HDU extension requested by the user does not exist."""


class HeaderFormatter:
    """Class to format the header(s) of a FITS file for display by the
    `fitsheader` tool; essentially a wrapper around a `HDUList` object.

    Example usage:
    fmt = HeaderFormatter('/path/to/file.fits')
    print(fmt.parse(extensions=[0, 3], keywords=['NAXIS', 'BITPIX']))

    Parameters
    ----------
    filename : str
        Path to a single FITS file.
    verbose : bool
        Verbose flag, to show more information about missing extensions,
        keywords, etc.

    Raises
    ------
    OSError
        If `filename` does not exist or cannot be read.
    """

    def __init__(self, filename, verbose=True):
        self.filename = filename
        self.verbose = verbose
        self._hdulist = fits.open(filename)

    def parse(self, extensions=None, keywords=None, compressed=False):
        """Returns the FITS file header(s) in a readable format.

        Parameters
        ----------
        extensions : list of int or str, optional
            Format only specific HDU(s), identified by number or name.
            The name can be composed of the "EXTNAME" or "EXTNAME,EXTVER"
            keywords.

        keywords : list of str, optional
            Keywords for which the value(s) should be returned.
            If not specified, then the entire header is returned.

        compressed : bool, optional
            If True, shows the header describing the compression, rather than
            the header obtained after decompression. (Affects FITS files
            containing `CompImageHDU` extensions only.)

        Returns
        -------
        formatted_header : str or astropy.table.Table
            Traditional 80-char wide format in the case of `HeaderFormatter`;
            an Astropy Table object in the case of `TableHeaderFormatter`.
        """
        # `hdukeys` will hold the keys of the HDUList items to display
        if extensions is None:
            hdukeys = range(len(self._hdulist))  # Display all by default
        else:
            hdukeys = []
            for ext in extensions:
                try:
                    # HDU may be specified by number
                    hdukeys.append(int(ext))
                except ValueError:
                    # The user can specify "EXTNAME" or "EXTNAME,EXTVER"
                    parts = ext.split(",")
                    if len(parts) > 1:
                        extname = ",".join(parts[0:-1])
                        extver = int(parts[-1])
                        hdukeys.append((extname, extver))
                    else:
                        hdukeys.append(ext)

        # Having established which HDUs the user wants, we now format these:
        return self._parse_internal(hdukeys, keywords, compressed)

    def _parse_internal(self, hdukeys, keywords, compressed):
        """The meat of the formatting; in a separate method to allow overriding."""
        result = []
        for idx, hdu in enumerate(hdukeys):
            try:
                cards = self._get_cards(hdu, keywords, compressed)
            except ExtensionNotFoundException:
                continue

            if idx > 0:  # Separate HDUs by a blank line
                result.append("\n")
            result.append(f"# HDU {hdu} in {self.filename}:\n")
            for c in cards:
                result.append(f"{c}\n")
        return "".join(result)

    def _get_cards(self, hdukey, keywords, compressed):
        """Returns a list of `astropy.io.fits.card.Card` objects.

        This function will return the desired header cards, taking into
        account the user's preference to see the compressed or uncompressed
        version.

        Parameters
        ----------
        hdukey : int or str
            Key of a single HDU in the HDUList.

        keywords : list of str, optional
            Keywords for which the cards should be returned.

        compressed : bool, optional
            If True, shows the header describing the compression.

        Raises
        ------
        ExtensionNotFoundException
            If the hdukey does not correspond to an extension.
        """
        # First we obtain the desired header
        try:
            if compressed and isinstance(self._hdulist[hdukey], CompImageHDU):
                # In the case of a compressed image, return the header before
                # decompression (not the default behavior)
                header = self._hdulist[hdukey]._bintable.header
            else:
                header = self._hdulist[hdukey].header
        except (IndexError, KeyError):
            message = f"{self.filename}: Extension {hdukey} not found."
            if self.verbose:
                log.warning(message)
            raise ExtensionNotFoundException(message)

        if not keywords:  # return all cards
            cards = header.cards
        else:  # specific keywords are requested
            cards = []
            for kw in keywords:
                try:
                    crd = header.cards[kw]
                    if isinstance(crd, fits.card.Card):  # Single card
                        cards.append(crd)
                    else:  # Allow for wildcard access
                        cards.extend(crd)
                except KeyError:  # Keyword does not exist
                    if self.verbose:
                        log.warning(
                            f"{self.filename} (HDU {hdukey}): Keyword {kw} not found."
                        )
        return cards

    def close(self):
        self._hdulist.close()


class TableHeaderFormatter(HeaderFormatter):
    """Class to convert the header(s) of a FITS file into a Table object.
    The table returned by the `parse` method will contain four columns:
    filename, hdu, keyword, and value.

    Subclassed from HeaderFormatter, which contains the meat of the formatting.
    """

    def _parse_internal(self, hdukeys, keywords, compressed):
        """Method called by the parse method in the parent class."""
        tablerows = []
        for hdu in hdukeys:
            try:
                for card in self._get_cards(hdu, keywords, compressed):
                    tablerows.append(
                        {
                            "filename": self.filename,
                            "hdu": hdu,
                            "keyword": card.keyword,
                            "value": str(card.value),
                        }
                    )
            except ExtensionNotFoundException:
                pass

        if tablerows:
            from astropy import table

            return table.Table(tablerows)
        return None


def print_headers_traditional(args):
    """Prints FITS header(s) using the traditional 80-char format.

    Parameters
    ----------
    args : argparse.Namespace
        Arguments passed from the command-line as defined below.
    """
    for idx, filename in enumerate(args.filename):  # support wildcards
        if idx > 0 and not args.keyword:
            print()  # print a newline between different files

        formatter = None
        try:
            formatter = HeaderFormatter(filename)
            print(
                formatter.parse(args.extensions, args.keyword, args.compressed), end=""
            )
        except OSError as e:
            log.error(str(e))
        finally:
            if formatter:
                formatter.close()


def print_headers_as_table(args):
    """Prints FITS header(s) in a machine-readable table format.

    Parameters
    ----------
    args : argparse.Namespace
        Arguments passed from the command-line as defined below.
    """
    tables = []
    # Create a Table object for each file
    for filename in args.filename:  # Support wildcards
        formatter = None
        try:
            formatter = TableHeaderFormatter(filename)
            tbl = formatter.parse(args.extensions, args.keyword, args.compressed)
            if tbl:
                tables.append(tbl)
        except OSError as e:
            log.error(str(e))  # file not found or unreadable
        finally:
            if formatter:
                formatter.close()

    # Concatenate the tables
    if len(tables) == 0:
        return False
    elif len(tables) == 1:
        resulting_table = tables[0]
    else:
        from astropy import table

        resulting_table = table.vstack(tables)
    # Print the string representation of the concatenated table
    resulting_table.write(sys.stdout, format=args.table)


def print_headers_as_comparison(args):
    """Prints FITS header(s) with keywords as columns.

    This follows the dfits+fitsort format.

    Parameters
    ----------
    args : argparse.Namespace
        Arguments passed from the command-line as defined below.
    """
    from astropy import table

    tables = []
    # Create a Table object for each file
    for filename in args.filename:  # Support wildcards
        formatter = None
        try:
            formatter = TableHeaderFormatter(filename, verbose=False)
            tbl = formatter.parse(args.extensions, args.keyword, args.compressed)
            if tbl:
                # Remove empty keywords
                tbl = tbl[np.where(tbl["keyword"] != "")]
            else:
                tbl = table.Table([[filename]], names=("filename",))
            tables.append(tbl)
        except OSError as e:
            log.error(str(e))  # file not found or unreadable
        finally:
            if formatter:
                formatter.close()

    # Concatenate the tables
    if len(tables) == 0:
        return False
    elif len(tables) == 1:
        resulting_table = tables[0]
    else:
        resulting_table = table.vstack(tables)

    # If we obtained more than one hdu, merge hdu and keywords columns
    hdus = resulting_table["hdu"]
    if np.ma.isMaskedArray(hdus):
        hdus = hdus.compressed()
    if len(np.unique(hdus)) > 1:
        for tab in tables:
            new_column = table.Column([f"{row['hdu']}:{row['keyword']}" for row in tab])
            tab.add_column(new_column, name="hdu+keyword")
        keyword_column_name = "hdu+keyword"
    else:
        keyword_column_name = "keyword"

    # Check how many hdus we are processing
    final_tables = []
    for tab in tables:
        final_table = [table.Column([tab["filename"][0]], name="filename")]
        if "value" in tab.colnames:
            for row in tab:
                if row["keyword"] in ("COMMENT", "HISTORY"):
                    continue
                final_table.append(
                    table.Column([row["value"]], name=row[keyword_column_name])
                )
        final_tables.append(table.Table(final_table))
    final_table = table.vstack(final_tables)
    # Sort if requested
    if args.sort:
        final_table.sort(args.sort)
    # Reorganise to keyword by columns
    final_table.pprint(max_lines=-1, max_width=-1)


def main(args=None):
    """This is the main function called by the `fitsheader` script."""
    parser = argparse.ArgumentParser(
        description=DESCRIPTION, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    # TODO: pass suggest_on_error as kwarg when PYTHON_LT_14 is dropped
    parser.suggest_on_error = True

    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )

    parser.add_argument(
        "-e",
        "--extension",
        metavar="HDU",
        action="append",
        dest="extensions",
        help=(
            "specify the extension by name or number; this argument can "
            "be repeated to select multiple extensions"
        ),
    )
    parser.add_argument(
        "-k",
        "--keyword",
        metavar="KEYWORD",
        action="append",
        type=str,
        help=(
            "specify a keyword; this argument can be repeated to select "
            "multiple keywords; also supports wildcards"
        ),
    )
    mode_group = parser.add_mutually_exclusive_group()
    mode_group.add_argument(
        "-t",
        "--table",
        nargs="?",
        default=False,
        metavar="FORMAT",
        help=(
            "print the header(s) in machine-readable table format; the "
            'default format is "ascii.fixed_width" (can be "ascii.csv", '
            '"ascii.html", "ascii.latex", "fits", etc)'
        ),
    )
    mode_group.add_argument(
        "-f",
        "--fitsort",
        action="store_true",
        help=(
            "print the headers as a table with each unique "
            "keyword in a given column (fitsort format) "
        ),
    )
    parser.add_argument(
        "-s",
        "--sort",
        metavar="SORT_KEYWORD",
        action="append",
        type=str,
        help=(
            "sort output by the specified header keywords, can be repeated to "
            "sort by multiple keywords; Only supported with -f/--fitsort"
        ),
    )
    parser.add_argument(
        "-c",
        "--compressed",
        action="store_true",
        help=(
            "for compressed image data, show the true header which describes "
            "the compression rather than the data"
        ),
    )
    parser.add_argument(
        "filename",
        nargs="+",
        help="path to one or more files; wildcards are supported",
    )
    args = parser.parse_args(args)

    # If `--table` was used but no format specified,
    # then use ascii.fixed_width by default
    if args.table is None:
        args.table = "ascii.fixed_width"

    if args.sort:
        args.sort = [key.replace(".", " ") for key in args.sort]
        if not args.fitsort:
            log.error(
                "Sorting with -s/--sort is only supported in conjunction with"
                " -f/--fitsort"
            )
            # 2: Unix error convention for command line syntax
            sys.exit(2)

    if args.keyword:
        args.keyword = [key.replace(".", " ") for key in args.keyword]

    # Now print the desired headers
    try:
        if args.table:
            print_headers_as_table(args)
        elif args.fitsort:
            print_headers_as_comparison(args)
        else:
            print_headers_traditional(args)
    except OSError:
        # A 'Broken pipe' OSError may occur when stdout is closed prematurely,
        # eg. when calling `fitsheader file.fits | head`. We let this pass.
        pass



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Helper functions for table development, mostly creating useful
tables for testing.
"""

import string
import warnings
from itertools import cycle

import numpy as np

from astropy.io.votable.table import parse
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.data_info import ParentDtypeInfo

from .table import Column, Table


def simple_table(size=3, cols=None, kinds="ifS", masked=False):
    """
    Return a simple table for testing.

    Example
    --------
    ::

      >>> from astropy.table.table_helpers import simple_table
      >>> print(simple_table(3, 6, masked=True, kinds='ifOS'))
       a   b     c      d   e   f
      --- --- -------- --- --- ---
       -- 1.0 {'c': 2}  --   5 5.0
        2 2.0       --   e   6  --
        3  -- {'e': 4}   f  -- 7.0

    Parameters
    ----------
    size : int
        Number of table rows
    cols : int, optional
        Number of table columns. Defaults to number of kinds.
    kinds : str
        String consisting of the column dtype.kinds.  This string
        will be cycled through to generate the column dtype.
        The allowed values are 'i', 'f', 'S', 'O'.

    Returns
    -------
    out : `Table`
        New table with appropriate characteristics
    """
    if cols is None:
        cols = len(kinds)
    if cols > 26:
        raise ValueError("Max 26 columns in SimpleTable")

    columns = []
    names = [chr(ord("a") + ii) for ii in range(cols)]
    letters = np.array(list(string.ascii_letters))
    for jj, kind in zip(range(cols), cycle(kinds)):
        if kind == "i":
            data = np.arange(1, size + 1, dtype=np.int64) + jj
        elif kind == "f":
            data = np.arange(size, dtype=np.float64) + jj
        elif kind == "S":
            indices = (np.arange(size) + jj) % len(letters)
            data = letters[indices]
        elif kind == "O":
            indices = (np.arange(size) + jj) % len(letters)
            vals = letters[indices]
            data = [{val.item(): index.item()} for val, index in zip(vals, indices)]
        else:
            raise ValueError("Unknown data kind")
        columns.append(Column(data))

    table = Table(columns, names=names, masked=masked)
    if masked:
        for ii, col in enumerate(table.columns.values()):
            mask = np.array((np.arange(size) + ii) % 3, dtype=bool)
            col.mask = ~mask

    return table


def complex_table():
    """
    Return a masked table from the io.votable test set that has a wide variety
    of stressing types.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        votable = parse(
            get_pkg_data_filename("../io/votable/tests/data/regression.xml"),
            pedantic=False,
        )
    first_table = votable.get_first_table()
    table = first_table.to_table()

    return table


class ArrayWrapperInfo(ParentDtypeInfo):
    _represent_as_dict_primary_data = "data"

    def _represent_as_dict(self):
        """Represent Column as a dict that can be serialized."""
        col = self._parent
        out = {"data": col.data}
        return out

    def _construct_from_dict(self, map):
        """Construct Column from ``map``."""
        data = map.pop("data")
        out = self._parent_cls(data, **map)
        return out


class ArrayWrapper:
    """
    Minimal mixin using a simple wrapper around a numpy array.

    TODO: think about the future of this class as it is mostly for demonstration
    purposes (of the mixin protocol). Consider taking it out of core and putting
    it into a tutorial. One advantage of having this in core is that it is
    getting tested in the mixin testing though it doesn't work for multidim
    data.
    """

    info = ArrayWrapperInfo()

    def __init__(self, data, copy=True):
        if isinstance(data, ArrayWrapper):
            # this is done to preserve byteorder through copies
            arr = data.data
        else:
            arr = data
        self.data = np.array(arr, copy=copy)
        if "info" in getattr(data, "__dict__", ()):
            self.info = data.info

    def __getitem__(self, item):
        if isinstance(item, (int, np.integer)):
            out = self.data[item]
        else:
            out = self.__class__(self.data[item], copy=False)
            if "info" in self.__dict__:
                out.info = self.info
        return out

    def __setitem__(self, item, value):
        self.data[item] = value

    def __len__(self):
        return len(self.data)

    def __eq__(self, other):
        """Minimal equality testing, mostly for mixin unit tests."""
        if isinstance(other, ArrayWrapper):
            return self.data == other.data
        else:
            return self.data == other

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def shape(self):
        return self.data.shape

    def __repr__(self):
        return f"<{self.__class__.__name__} name='{self.info.name}' data={self.data}>"



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
from pathlib import Path

__all__ = [
    "setup_function",
    "teardown_function",
]

CWD = Path.cwd()
TEST_DIR = Path(__file__).parent


def setup_function(function):
    os.chdir(TEST_DIR)


def teardown_function(function):
    os.chdir(CWD)


def assert_equal_splitlines(arg1, arg2):
    __tracebackhide__ = True
    assert arg1.splitlines() == arg2.splitlines()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This package defines units used in the CDS format, both the units
defined in `Centre de Données astronomiques de Strasbourg
<https://cds.unistra.fr/>`_ `Standards for Astronomical Catalogues 2.0
<https://vizier.unistra.fr/vizier/doc/catstd-3.2.htx>`_ format and the `complete
set of supported units <https://vizier.unistra.fr/viz-bin/Unit>`_.
This format is used by VOTable up to version 1.2.

These units are not available in the top-level `astropy.units`
namespace.  To use these units, you must import the `astropy.units.cds`
module::

    >>> from astropy.units import cds
    >>> q = 10. * cds.lyr  # doctest: +SKIP

To include them in `~astropy.units.UnitBase.compose` and the results of
`~astropy.units.UnitBase.find_equivalent_units`, do::

    >>> from astropy.units import cds
    >>> cds.enable()  # doctest: +SKIP

Note, however, that this can introduce conflicts between CDS and other
units in the top-level namespace. A safer way to use CDS units is to enable
them inside a context manager. For instance, you could do the following if
you have a string that uses CDS units:

>>> import astropy.units as u
>>> unit_string = "mmHg"
>>> with cds.enable():
...     pressure_unit = u.Unit(unit_string)
>>> (720*pressure_unit).to(u.bar)
<Quantity 0.95992119 bar>
"""

__all__ = ["enable"]  #  Units are added at the end

import numpy as np

from astropy.constants import si as _si

from .core import (
    _UnitContext,
    binary_prefixes,
    def_unit,
    set_enabled_units,
    si_prefixes,
)
from .docgen import generate_dunder_all, generate_unit_summary

_ns = globals()


def _initialize_module():
    # Having `u` in the global namespace would conflict with the atomic mass unit.
    from astropy import units as u

    prefixes = []
    # The CDS format also supports power-of-2 prefixes as defined here:
    # http://physics.nist.gov/cuu/Units/binary.html
    for short, _, factor in si_prefixes + binary_prefixes:
        short = [s for s in short if s.isascii()]
        prefixes.append((short, short, factor))  # CDS only uses the short prefixes

    # The following units are defined in alphabetical order, directly from
    # here: https://vizier.unistra.fr/viz-bin/Unit

    mapping = [
        (["A"], u.A, "Ampere"),
        (["a"], u.a, "year", ["P"]),
        (["a0"], _si.a0, "Bohr radius"),
        (["al"], u.lyr, "Light year", ["c", "d"]),
        (["lyr"], u.lyr, "Light year"),
        (["alpha"], _si.alpha, "Fine structure constant"),
        (["Angstrom", "Å", "Angstroem", "AA"], u.AA, "Angstrom"),
        (["arcmin", "arcm"], u.arcminute, "minute of arc"),
        (["arcsec", "arcs"], u.arcsecond, "second of arc"),
        (["atm"], _si.atm, "atmosphere"),
        (["AU", "au"], u.au, "astronomical unit"),
        (["bar"], u.bar, "bar"),
        (["barn"], u.barn, "barn"),
        (["bit"], u.bit, "bit"),
        (["byte"], u.byte, "byte"),
        (["C"], u.C, "Coulomb"),
        (["c"], _si.c, "speed of light", ["p"]),
        (["cal"], 4.1854 * u.J, "calorie"),
        (["cd"], u.cd, "candela"),
        (["ct"], u.ct, "count"),
        (["D"], u.D, "Debye (dipole)"),
        (["d"], u.d, "Julian day", ["c"]),
        ((["deg", "°"], ["degree"]), u.degree, "degree"),
        (["dyn"], u.dyn, "dyne"),
        (["e"], _si.e, "electron charge", ["m"]),
        (["eps0"], _si.eps0, "electric constant"),
        (["erg"], u.erg, "erg"),
        (["eV"], u.eV, "electron volt"),
        (["F"], u.F, "Farad"),
        (["G"], _si.G, "Gravitation constant"),
        (["g"], u.g, "gram"),
        (["gauss"], u.G, "Gauss"),
        (["geoMass", "Mgeo"], u.M_earth, "Earth mass"),
        (["H"], u.H, "Henry"),
        (["h"], u.h, "hour", ["p"]),
        (["hr"], u.h, "hour"),
        (["\\h"], _si.h, "Planck constant"),
        (["Hz"], u.Hz, "Hertz"),
        (["inch"], 0.0254 * u.m, "inch"),
        (["J"], u.J, "Joule"),
        (["JD"], u.d, "Julian day", ["M"]),
        (["jovMass", "Mjup"], u.M_jup, "Jupiter mass"),
        (["Jy"], u.Jy, "Jansky"),
        (["K"], u.K, "Kelvin"),
        (["k"], _si.k_B, "Boltzmann"),
        (["l"], u.l, "litre", ["a"]),
        (["lm"], u.lm, "lumen"),
        (["Lsun", "solLum"], u.solLum, "solar luminosity"),
        (["lx"], u.lx, "lux"),
        (["m"], u.m, "meter"),
        (["mag"], u.mag, "magnitude"),
        (["me"], _si.m_e, "electron mass"),
        (["min"], u.minute, "minute"),
        (["MJD"], u.d, "Julian day"),
        (["mmHg"], 133.322387415 * u.Pa, "millimeter of mercury"),
        (["mol"], u.mol, "mole"),
        (["mp"], _si.m_p, "proton mass"),
        (["solMass", "Msun"], u.solMass, "solar mass"),
        ((["mu0", "µ0"], []), _si.mu0, "magnetic constant"),
        (["muB"], _si.muB, "Bohr magneton"),
        (["N"], u.N, "Newton"),
        (["Ohm"], u.Ohm, "Ohm"),
        (["Pa"], u.Pa, "Pascal"),
        (["pc"], u.pc, "parsec"),
        (["ph"], u.ph, "photon"),
        (["pi"], u.Unit(np.pi), "π"),
        (["pix"], u.pix, "pixel"),
        (["ppm"], u.Unit(1e-6), "parts per million"),
        (["R"], _si.R, "gas constant"),
        (["rad"], u.radian, "radian"),
        (["Rgeo"], _si.R_earth, "Earth equatorial radius"),
        (["Rjup"], _si.R_jup, "Jupiter equatorial radius"),
        (["Rsun", "solRad"], u.solRad, "solar radius"),
        (["Ry"], u.Ry, "Rydberg"),
        (["S"], u.S, "Siemens"),
        (["s", "sec"], u.s, "second"),
        (["sr"], u.sr, "steradian"),
        (["Sun"], u.Sun, "solar unit"),
        (["T"], u.T, "Tesla"),
        (["t"], 1e3 * u.kg, "metric tonne", ["c"]),
        (["u"], _si.u, "atomic mass", ["da", "a"]),
        (["V"], u.V, "Volt"),
        (["W"], u.W, "Watt"),
        (["Wb"], u.Wb, "Weber"),
        (["yr"], u.a, "year"),
    ]

    for entry in mapping:
        if len(entry) == 3:
            names, unit, doc = entry
            excludes = []
        else:
            names, unit, doc, excludes = entry
        def_unit(
            names,
            unit,
            prefixes=prefixes,
            namespace=_ns,
            doc=doc,
            exclude_prefixes=excludes,
        )

    def_unit(["µas"], u.microarcsecond, doc="microsecond of arc", namespace=_ns)
    def_unit(["mas"], u.milliarcsecond, doc="millisecond of arc", namespace=_ns)
    def_unit(
        ["---", "-"],
        u.dimensionless_unscaled,
        doc="dimensionless and unscaled",
        namespace=_ns,
    )
    def_unit(["%"], u.percent, doc="percent", namespace=_ns)
    # The Vizier "standard" defines this in units of "kg s-3", but
    # that may not make a whole lot of sense, so here we just define
    # it as its own new disconnected unit.
    def_unit(["Crab"], prefixes=prefixes, namespace=_ns, doc="Crab (X-ray) flux")


_initialize_module()


###########################################################################
# ALL & DOCSTRING

__all__ += generate_dunder_all(globals())  # noqa: PLE0605

if __doc__ is not None:
    # This generates a docstring for this module that describes all of the
    # standard units defined here.
    __doc__ += generate_unit_summary(globals())


def enable() -> _UnitContext:
    """
    Enable CDS units so they appear in results of
    `~astropy.units.UnitBase.find_equivalent_units` and
    `~astropy.units.UnitBase.compose`.  This will disable
    all of the "default" `astropy.units` units, since there
    are some namespace clashes between the two.

    This may be used with the ``with`` statement to enable CDS
    units only temporarily.
    """
    return set_enabled_units(globals())



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest

import astropy.units as u
from astropy.cosmology import Cosmology, FlatLambdaCDM, Planck18
from astropy.cosmology import units as cu
from astropy.cosmology._src.io.builtin.yaml import (
    from_yaml,
    to_yaml,
    yaml_constructor,
    yaml_representer,
)
from astropy.io.misc.yaml import AstropyDumper, dump, load

from .base import ToFromDirectTestBase, ToFromTestMixinBase

##############################################################################
# Test Serializer


def test_yaml_representer():
    """Test :func:`~astropy.cosmology._src.io.builtin.yaml.yaml_representer`."""
    # test function `representer`
    representer = yaml_representer("!astropy.cosmology.LambdaCDM")
    assert callable(representer)

    # test the normal method of dumping to YAML
    yml = dump(Planck18)
    assert isinstance(yml, str)
    assert yml.startswith("!astropy.cosmology.FlatLambdaCDM")


def test_yaml_constructor():
    """Test :func:`~astropy.cosmology._src.io.builtin.yaml.yaml_constructor`."""
    # test function `constructor`
    constructor = yaml_constructor(FlatLambdaCDM)
    assert callable(constructor)

    # it's too hard to manually construct a node, so we only test dump/load
    # this is also a good round-trip test
    yml = dump(Planck18)
    with u.add_enabled_units(cu):  # needed for redshift units
        cosmo = load(yml)
    assert isinstance(cosmo, FlatLambdaCDM)
    assert cosmo == Planck18
    assert cosmo.meta == Planck18.meta


##############################################################################
# Test Unified I/O


class ToFromYAMLTestMixin(ToFromTestMixinBase):
    """
    Tests for a Cosmology[To/From]Format with ``format="yaml"``.
    This class will not be directly called by :mod:`pytest` since its name does
    not begin with ``Test``. To activate the contained tests this class must
    be inherited in a subclass. Subclasses must define a :func:`pytest.fixture`
    ``cosmo`` that returns/yields an instance of a |Cosmology|.
    See ``TestCosmologyToFromFormat`` or ``TestCosmology`` for examples.
    """

    @pytest.fixture
    def xfail_if_not_registered_with_yaml(self, cosmo_cls):
        """
        YAML I/O only works on registered classes. So the thing to check is
        if this class is registered. If not, :func:`pytest.xfail` this test.
        Some of the tests define custom cosmologies. They are not registered.
        """
        if cosmo_cls not in AstropyDumper.yaml_representers:
            pytest.xfail(
                f"Cosmologies of type {cosmo_cls} are not registered with YAML."
            )

    # ===============================================================

    def test_to_yaml(self, cosmo_cls, to_format, xfail_if_not_registered_with_yaml):
        """Test cosmology -> YAML."""
        yml = to_format("yaml")

        assert isinstance(yml, str)  # test type
        assert yml.startswith("!" + ".".join(cosmo_cls.__module__.split(".")[:2]))
        # e.g. "astropy.cosmology" for built-in cosmologies, or "__main__" for the test
        # SubCosmology class defined in ``astropy.cosmology._src.tests.test_core``.

    def test_from_yaml_default(
        self, cosmo, to_format, from_format, xfail_if_not_registered_with_yaml
    ):
        """Test cosmology -> YAML -> cosmology."""
        yml = to_format("yaml")

        got = from_format(yml, format="yaml")  # (cannot autoidentify)

        assert got.name == cosmo.name
        assert got.meta == cosmo.meta

        # it won't error if everything matches up
        got = from_format(yml, format="yaml")
        assert got == cosmo
        assert got.meta == cosmo.meta

        # auto-identify test moved because it doesn't work.
        # see test_from_yaml_autoidentify

    def test_from_yaml_autoidentify(
        self, cosmo, to_format, from_format, xfail_if_not_registered_with_yaml
    ):
        """As a non-path string, it does NOT auto-identifies 'format'.

        TODO! this says there should be different types of I/O registries.
              not just hacking object conversion on top of file I/O.
        """
        assert self.can_autodentify("yaml") is False

        # Showing the specific error. The str is interpreted as a file location
        # but is too long a file name.
        yml = to_format("yaml")
        with pytest.raises((FileNotFoundError, OSError)):  # OSError in Windows
            from_format(yml)

    # # TODO! this is a challenging test to write. It's also unlikely to happen.
    # def test_fromformat_subclass_partial_info_yaml(self, cosmo):
    #     """
    #     Test writing from an instance and reading from that class.
    #     This works with missing information.
    #     """

    # -----------------------------------------------------

    @pytest.mark.parametrize("format", [True, False, None])
    def test_is_equivalent_to_yaml(
        self, cosmo, to_format, format, xfail_if_not_registered_with_yaml
    ):
        """Test :meth:`astropy.cosmology.Cosmology.is_equivalent`.

        This test checks that Cosmology equivalency can be extended to any
        Python object that can be converted to a Cosmology -- in this case
        a YAML string. YAML can't be identified without "format" specified.
        """
        obj = to_format("yaml")
        assert not isinstance(obj, Cosmology)

        is_equiv = cosmo.is_equivalent(obj, format=format)
        assert is_equiv is False

    def test_is_equivalent_to_yaml_specify_format(
        self, cosmo, to_format, xfail_if_not_registered_with_yaml
    ):
        """Test :meth:`astropy.cosmology.Cosmology.is_equivalent`.

        Same as ``test_is_equivalent_to_yaml`` but with ``format="yaml"``.
        """
        assert cosmo.is_equivalent(to_format("yaml"), format="yaml") is True


class TestToFromYAML(ToFromDirectTestBase, ToFromYAMLTestMixin):
    """
    Directly test ``to/from_yaml``.
    These are not public API and are discouraged from use, in favor of
    ``Cosmology.to/from_format(..., format="yaml")``, but should be tested
    regardless b/c 3rd party packages might use these in their Cosmology I/O.
    Also, it's cheap to test.
    """

    def setup_class(self):
        """Set up fixtures to use ``to/from_yaml``, not the I/O abstractions."""
        self.functions = {"to": to_yaml, "from": from_yaml}

    @pytest.fixture(scope="class", autouse=True)
    def setup(self):
        """
        Setup and teardown for tests.
        This overrides from super because `ToFromDirectTestBase` adds a custom
        Cosmology ``CosmologyWithKwargs`` that is not registered with YAML.
        """
        return  # run tests

    def test_from_yaml_autoidentify(self, cosmo, to_format, from_format):
        """
        If directly calling the function there's no auto-identification.
        So this overrides the test from `ToFromYAMLTestMixin`
        """



import pytest

from astropy.tests.helper import _skip_docstring_tests_with_optimized_python

# Renamed these imports so that them being in the namespace will not
# cause pytest 3 to discover them as tests and then complain that
# they have __init__ defined.
from astropy.tests.runner import TestRunner as _TestRunner
from astropy.tests.runner import TestRunnerBase as _TestRunnerBase
from astropy.tests.runner import keyword


def test_disable_kwarg():
    class no_remote_data(_TestRunner):
        @keyword()
        def remote_data(self, remote_data, kwargs):
            return NotImplemented

    with pytest.deprecated_call(match="The TestRunner"):
        r = no_remote_data(".")
    with pytest.raises(TypeError), pytest.deprecated_call(match="The test runner"):
        r.run_tests(remote_data="bob")


def test_wrong_kwarg():
    with pytest.deprecated_call(match="The TestRunner"):
        r = _TestRunner(".")
    with pytest.raises(TypeError), pytest.deprecated_call(match="The test runner"):
        r.run_tests(spam="eggs")


def test_invalid_kwarg():
    class bad_return(_TestRunnerBase):
        @keyword()
        def remote_data(self, remote_data, kwargs):
            return "bob"

    with pytest.deprecated_call(match="The TestRunner"):
        r = bad_return(".")
    with pytest.raises(TypeError), pytest.deprecated_call(match="The test runner"):
        r.run_tests(remote_data="bob")


def test_new_kwarg():
    class Spam(_TestRunnerBase):
        @keyword()
        def spam(self, spam, kwargs):
            return [spam]

    with pytest.deprecated_call(match="The TestRunner"):
        r = Spam(".")

    args = r._generate_args(spam="spam")

    assert ["spam"] == args


def test_priority():
    class Spam(_TestRunnerBase):
        @keyword()
        def spam(self, spam, kwargs):
            return [spam]

        @keyword(priority=1)
        def eggs(self, eggs, kwargs):
            return [eggs]

    with pytest.deprecated_call(match="The TestRunner"):
        r = Spam(".")

    args = r._generate_args(spam="spam", eggs="eggs")

    assert ["eggs", "spam"] == args


@_skip_docstring_tests_with_optimized_python
def test_docs():
    class Spam(_TestRunnerBase):
        @keyword()
        def spam(self, spam, kwargs):
            """
            Spam Spam Spam
            """
            return [spam]

        @keyword()
        def eggs(self, eggs, kwargs):
            """
            eggs asldjasljd
            """
            return [eggs]

    with pytest.deprecated_call(match="The TestRunner"):
        r = Spam(".")
    assert "deprecated" in r.run_tests.__doc__
    assert "eggs" in r.run_tests.__doc__
    assert "Spam Spam Spam" in r.run_tests.__doc__



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest

from astropy import units as u
from astropy.coordinates import EarthLocation, SkyCoord, solar_system_ephemeris
from astropy.time import Time, TimeDelta
from astropy.utils import iers
from astropy.utils.compat.optional_deps import HAS_JPLEPHEM


class TestHelioBaryCentric:
    """
    Verify time offsets to the solar system barycentre and the heliocentre.
    Uses the WHT observing site.

    Tests are against values returned at time of initial creation of these
    routines.  They agree to an independent SLALIB based implementation
    to 20 microseconds.
    """

    @classmethod
    def setup_class(cls):
        cls.orig_auto_download = iers.conf.auto_download
        iers.conf.auto_download = False

    @classmethod
    def teardown_class(cls):
        iers.conf.auto_download = cls.orig_auto_download

    def setup_method(self):
        wht = EarthLocation(342.12 * u.deg, 28.758333333333333 * u.deg, 2327 * u.m)
        self.obstime = Time("2013-02-02T23:00", location=wht)
        self.obstime2 = Time("2013-08-02T23:00", location=wht)
        self.obstimeArr = Time(["2013-02-02T23:00", "2013-08-02T23:00"], location=wht)
        self.star = SkyCoord(
            "08:08:08 +32:00:00", unit=(u.hour, u.degree), frame="icrs"
        )

    def test_heliocentric(self):
        hval = self.obstime.light_travel_time(self.star, "heliocentric")
        assert isinstance(hval, TimeDelta)
        assert hval.scale == "tdb"
        assert abs(hval - 461.43037870502235 * u.s) < 1.0 * u.us

    def test_barycentric(self):
        bval = self.obstime.light_travel_time(self.star, "barycentric")
        assert isinstance(bval, TimeDelta)
        assert bval.scale == "tdb"
        assert abs(bval - 460.58538779827836 * u.s) < 1.0 * u.us

    def test_arrays(self):
        bval1 = self.obstime.light_travel_time(self.star, "barycentric")
        bval2 = self.obstime2.light_travel_time(self.star, "barycentric")
        bval_arr = self.obstimeArr.light_travel_time(self.star, "barycentric")
        hval1 = self.obstime.light_travel_time(self.star, "heliocentric")
        hval2 = self.obstime2.light_travel_time(self.star, "heliocentric")
        hval_arr = self.obstimeArr.light_travel_time(self.star, "heliocentric")
        assert hval_arr[0] - hval1 < 1.0 * u.us
        assert hval_arr[1] - hval2 < 1.0 * u.us
        assert bval_arr[0] - bval1 < 1.0 * u.us
        assert bval_arr[1] - bval2 < 1.0 * u.us

    @pytest.mark.remote_data
    @pytest.mark.skipif(not HAS_JPLEPHEM, reason="requires jplephem")
    def test_ephemerides(self):
        bval1 = self.obstime.light_travel_time(self.star, "barycentric")
        with solar_system_ephemeris.set("jpl"):
            bval2 = self.obstime.light_travel_time(
                self.star, "barycentric", ephemeris="jpl"
            )
        # should differ by less than 0.1 ms, but not be the same
        assert abs(bval1 - bval2) < 1.0 * u.ms
        assert abs(bval1 - bval2) > 1.0 * u.us



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import filecmp
import io

import numpy as np
import pytest

from astropy.io.votable import tree
from astropy.io.votable.exceptions import (
    E26,
    W07,
    W08,
    W15,
    W21,
    W27,
    W37,
    W41,
    W47,
    W54,
    W57,
)
from astropy.io.votable.table import parse
from astropy.io.votable.tree import MivotBlock, Resource, VOTableFile
from astropy.utils.data import get_pkg_data_filename


def test_check_astroyear_fail():
    config = {"verify": "exception"}
    field = tree.Field(None, name="astroyear", arraysize="1")
    with pytest.raises(W07):
        tree.check_astroyear("X2100", field, config)


def test_string_fail():
    config = {"verify": "exception"}
    with pytest.raises(W08):
        tree.check_string(42, "foo", config)


def test_make_Fields():
    votable = VOTableFile()
    # ...with one resource...
    resource = Resource()
    votable.resources.append(resource)

    # ... with one table
    table = tree.TableElement(votable)
    resource.tables.append(table)

    table.fields.extend(
        [tree.Field(votable, name="Test", datatype="float", unit="mag")]
    )


def test_unit_format():
    data = parse(get_pkg_data_filename("data/irsa-nph-error.xml"))
    assert data._config["version"] == "1.0"
    assert tree._get_default_unit_format(data._config) == "cds"
    data = parse(get_pkg_data_filename("data/names.xml"))
    assert data._config["version"] == "1.1"
    assert tree._get_default_unit_format(data._config) == "cds"
    data = parse(get_pkg_data_filename("data/gemini.xml"))
    assert data._config["version"] == "1.2"
    assert tree._get_default_unit_format(data._config) == "cds"
    data = parse(get_pkg_data_filename("data/binary2_masked_strings.xml"))
    assert data._config["version"] == "1.3"
    assert tree._get_default_unit_format(data._config) == "cds"
    data = parse(get_pkg_data_filename("data/timesys.xml"))
    assert data._config["version"] == "1.4"
    assert tree._get_default_unit_format(data._config) == "vounit"


def test_namespace_warning():
    """
    A version 1.4 VOTable must use the same namespace as 1.3.
    (see https://www.ivoa.net/documents/VOTable/20191021/REC-VOTable-1.4-20191021.html#ToC16).
    """
    bad_namespace = b"""<?xml version="1.0" encoding="utf-8"?>
        <VOTABLE version="1.4" xmlns="http://www.ivoa.net/xml/VOTable/v1.4"
                               xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
          <RESOURCE/>
        </VOTABLE>
    """
    with pytest.warns(W41):
        parse(io.BytesIO(bad_namespace), verify="exception")

    good_namespace_14 = b"""<?xml version="1.0" encoding="utf-8"?>
        <VOTABLE version="1.4" xmlns="http://www.ivoa.net/xml/VOTable/v1.3"
                               xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
          <RESOURCE/>
        </VOTABLE>
    """
    parse(io.BytesIO(good_namespace_14), verify="exception")

    good_namespace_13 = b"""<?xml version="1.0" encoding="utf-8"?>
        <VOTABLE version="1.3" xmlns="http://www.ivoa.net/xml/VOTable/v1.3"
                               xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
          <RESOURCE/>
        </VOTABLE>
    """
    parse(io.BytesIO(good_namespace_13), verify="exception")


def test_votable_values_empty_min_max():
    """Regression test for https://github.com/astropy/astropy/issues/16825"""
    with_empty_minmax = b"""<VOTABLE xmlns="http://www.ivoa.net/xml/VOTable/v1.3" version="1.4">
        <RESOURCE type="results">
          <TABLE name="main">
            <PARAM name="break" datatype="int" value=""/>
          <FIELD ID="hd" datatype="int" name="hd" ucd="meta.id;meta.main">
            <DESCRIPTION>HD number for this object</DESCRIPTION>
            <VALUES null="-2147483648">
              <MIN value=""/>
              <MAX value=""/>
            </VALUES>
          </FIELD>
          <DATA>
            <BINARY>
              <STREAM encoding="base64">AAMNIg==</STREAM>
            </BINARY>
          </DATA>
        </TABLE>
      </RESOURCE>
    </VOTABLE>
    """
    parse(io.BytesIO(with_empty_minmax), verify="exception")


def _assert_minmax_match(val, expected_val):
    """val can be number or numpy array; expected_val can be number or list"""
    if isinstance(val, np.ndarray):
        assert np.allclose(val, np.array(expected_val))
    else:
        assert val == expected_val


@pytest.mark.parametrize(
    ("testvals"),
    [
        {
            "dt": "float",
            "min": "1.2",
            "max": "3.4",
            "expected_min": 1.2,
            "expected_max": 3.4,
        },
        {
            "dt": "float",
            "min": "1.2 3.4",
            "max": "3.4 5.6",
            "expected_min": [1.2, 3.4],
            "expected_max": [3.4, 5.6],
        },
        {
            "dt": "double",
            "min": "1.2",
            "max": "3.4",
            "expected_min": 1.2,
            "expected_max": 3.4,
        },
        {
            "dt": "double",
            "min": "1.2 3.4",
            "max": "3.4 5.6",
            "expected_min": [1.2, 3.4],
            "expected_max": [3.4, 5.6],
        },
        {
            "dt": "unsignedByte",
            "min": "1",
            "max": "3",
            "expected_min": 1,
            "expected_max": 3,
        },
        {
            "dt": "unsignedByte",
            "min": "1 3",
            "max": "3 5",
            "expected_min": [1, 3],
            "expected_max": [3, 5],
        },
        {"dt": "short", "min": "1", "max": "3", "expected_min": 1, "expected_max": 3},
        {
            "dt": "short",
            "min": "1 3",
            "max": "3 5",
            "expected_min": [1, 3],
            "expected_max": [3, 5],
        },
        {"dt": "int", "min": "1", "max": "3", "expected_min": 1, "expected_max": 3},
        {
            "dt": "int",
            "min": "1 3",
            "max": "3 5",
            "expected_min": [1, 3],
            "expected_max": [3, 5],
        },
        {"dt": "long", "min": "1", "max": "3", "expected_min": 1, "expected_max": 3},
        {
            "dt": "long",
            "min": "1 3",
            "max": "3 5",
            "expected_min": [1, 3],
            "expected_max": [3, 5],
        },
    ],
)
def test_min_max_with_arrays(testvals):
    """When a FIELD/PARAM is an array type, ensure we accept (without warnings/exceptions)
    MIN and MAX values both as scalars (preferred except for some xtypes) or arrays."""
    minmax_template = """<VOTABLE xmlns="http://www.ivoa.net/xml/VOTable/v1.3" version="1.5">
        <RESOURCE type="results">
            <TABLE ID="result" name="result">
                <{param_or_field} arraysize="2" datatype="{dt}" name="param_or_field" xtype="interval">
                    <VALUES><MIN value="{min}"/><MAX value="{max}"/></VALUES>
                </{param_or_field}>
            </TABLE>
        </RESOURCE>
    </VOTABLE>
    """

    for param_or_field in ["PARAM", "FIELD"]:
        testvals["param_or_field"] = param_or_field
        xml_string = minmax_template.format(**testvals)
        xml_bytes = xml_string.encode("utf-8")

        # Parse without exceptions.
        vot = parse(io.BytesIO(xml_bytes), verify="exception")

        # Assert that the scalar or array values were parsed and stored as expected.
        param_or_field = vot.get_field_by_id_or_name("param_or_field")
        _assert_minmax_match(param_or_field.values.min, testvals["expected_min"])
        _assert_minmax_match(param_or_field.values.max, testvals["expected_max"])


def test_version():
    """
    VOTableFile.__init__ allows versions of '1.1' through '1.5'.
    VOTableFile.__init__ does not allow version of '1.0' anymore and now raises a ValueError as it does to other versions not supported.
    """
    # Exercise the checks in __init__
    for version in ("1.1", "1.2", "1.3", "1.4", "1.5"):
        VOTableFile(version=version)
    for version in ("0.9", "1.0", "1.6", "2.0"):
        with pytest.raises(
            ValueError, match=r"should be in \('1.1', '1.2', '1.3', '1.4', '1.5'\)."
        ):
            VOTableFile(version=version)

    # Exercise the checks in the setter
    vot = VOTableFile()
    for version in ("1.1", "1.2", "1.3", "1.4", "1.5"):
        vot.version = version
    for version in ("1.0", "1.6", "2.0"):
        with pytest.raises(
            ValueError,
            match=r"version should be in \('1.1', '1.2', '1.3', '1.4', '1.5'\).",
        ):
            vot.version = version

    # Exercise the checks in the parser.
    begin = b'<?xml version="1.0" encoding="utf-8"?><VOTABLE version="'
    middle = b'" xmlns="http://www.ivoa.net/xml/VOTable/v'
    end = (
        b'" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"><RESOURCE/></VOTABLE>'
    )

    # Valid versions
    for bversion in (b"1.1", b"1.2", b"1.3"):
        parse(
            io.BytesIO(begin + bversion + middle + bversion + end), verify="exception"
        )
    parse(io.BytesIO(begin + b"1.4" + middle + b"1.3" + end), verify="exception")
    parse(io.BytesIO(begin + b"1.5" + middle + b"1.3" + end), verify="exception")

    # Invalid versions
    for bversion in (b"1.0", b"2.0"):
        with pytest.warns(W21), pytest.warns(W41):
            parse(
                io.BytesIO(begin + bversion + middle + bversion + end),
                verify="exception",
            )


def votable_xml_string(version):
    votable_file = VOTableFile(version=version)
    votable_file.resources.append(Resource())

    xml_bytes = io.BytesIO()
    votable_file.to_xml(xml_bytes)
    xml_bytes.seek(0)
    bstring = xml_bytes.read()
    return bstring.decode("utf-8")


def test_votable_tag():
    xml = votable_xml_string("1.1")
    assert 'xmlns="http://www.ivoa.net/xml/VOTable/v1.1"' in xml
    assert 'xsi:noNamespaceSchemaLocation="http://www.ivoa.net/xml/VOTable/v1.1"' in xml

    xml = votable_xml_string("1.2")
    assert 'xmlns="http://www.ivoa.net/xml/VOTable/v1.2"' in xml
    assert 'xsi:noNamespaceSchemaLocation="http://www.ivoa.net/xml/VOTable/v1.2"' in xml

    xml = votable_xml_string("1.3")
    assert 'xmlns="http://www.ivoa.net/xml/VOTable/v1.3"' in xml
    assert 'xsi:schemaLocation="http://www.ivoa.net/xml/VOTable/v1.3 ' in xml
    assert 'http://www.ivoa.net/xml/VOTable/VOTable-1.3.xsd"' in xml

    xml = votable_xml_string("1.4")
    assert 'xmlns="http://www.ivoa.net/xml/VOTable/v1.3"' in xml
    assert 'xsi:schemaLocation="http://www.ivoa.net/xml/VOTable/v1.3 ' in xml
    assert 'http://www.ivoa.net/xml/VOTable/VOTable-1.4.xsd"' in xml

    xml = votable_xml_string("1.5")
    assert 'xmlns="http://www.ivoa.net/xml/VOTable/v1.3"' in xml
    assert 'xsi:schemaLocation="http://www.ivoa.net/xml/VOTable/v1.3 ' in xml
    assert 'http://www.ivoa.net/xml/VOTable/VOTable-1.5.xsd"' in xml


def _squash_xml(data):
    """
    Utility squashing XML fragment to easier their comparison
    This function is only used in the test module. It was more convenient for comparing the xml.
    """
    return data.replace(" ", "").replace("\n", "").replace('"', "").replace("'", "")


def test_mivot_constructor():
    """
    Construct a MIVOT block with wrong tag to test the expected exception
    """
    with pytest.raises(ValueError, match="not well-formed"):
        MivotBlock(
            """
            <VODML xmlns="http://www.ivoa.net/xml/mivot" >
               <REPORT status="OK">Unit test mivot block1</REPORT>
               <WRONG TAG>
               </GLOBALS>
            </VODML>
            """
        )


def test_mivot_readout():
    """
    Test the MIVOT block extraction from a file against a reference block stored in data
    """
    votable = parse(get_pkg_data_filename("data/mivot_annotated_table.xml"))

    ref_data = ""
    for resource in votable.resources:
        with open(
            get_pkg_data_filename("data/mivot_block_custom_datatype.xml")
        ) as reference:
            ref_data = reference.read()
            assert _squash_xml(ref_data) == _squash_xml(resource.mivot_block.content)
        assert len(resource.tables) == 1


def test_mivot_write():
    """
    Build a VOTable, put a MIVOT block in the first resource, checks it can be retrieved
    as well as the following table
    """

    mivot_block = MivotBlock(
        """
    <VODML xmlns="http://www.ivoa.net/xml/mivot" >
       <REPORT status="OK">
       Unit test mivot block1
       </REPORT>
       <GLOBALS>
       </GLOBALS>
    </VODML>
    """
    )
    vtf = VOTableFile()
    mivot_resource = Resource()
    mivot_resource.type = "meta"
    mivot_resource.mivot_block = mivot_block
    # pack the meta resource in a top level resource
    r1 = Resource()
    r1.type = "results"
    r1.resources.append(mivot_resource)
    vtf.resources.append(r1)
    # Push the VOTable in an IOSTream (emulates a disk saving)
    buff = io.BytesIO()
    vtf.to_xml(buff)

    # Read the IOStream (emulates a disk readout)
    buff.seek(0)
    vtf2 = parse(buff)
    assert len(vtf2.resources) == 1
    for resource in vtf2.resources:
        assert _squash_xml(mivot_block.content) == _squash_xml(
            resource.mivot_block.content
        )
        assert len(resource.tables) == 0


def test_mivot_write_after_table():
    """
    Build a VOTable, put a MIVOT block and a table in the first resource, checks it can be retrieved
    as well as the following table
    """

    mivot_block = MivotBlock(
        """
    <VODML xmlns="http://www.ivoa.net/xml/mivot" >
      <REPORT status="OK">Unit test mivot block1</REPORT>
      <GLOBALS>
      </GLOBALS>
    </VODML>
    """
    )
    vtf = VOTableFile()
    mivot_resource = Resource()
    mivot_resource.type = "meta"
    mivot_resource.mivot_block = mivot_block
    # pack the meta resource in a top level resource
    r1 = Resource()
    r1.type = "results"
    i1 = tree.Info(name="test_name", value="test_value")
    r1.infos.append(i1)
    r1.resources.append(mivot_resource)
    t1 = tree.TableElement(vtf)
    t1.name = "t1"
    r1.tables.append(t1)
    vtf.resources.append(r1)
    # Push the VOTable in an IOSTream (emulates a disk saving)
    buff = io.BytesIO()
    vtf.to_xml(buff)

    # Read the IOStream (emulates a disk readout)
    buff.seek(0)
    vtf2 = parse(buff)
    assert len(vtf2.resources) == 1
    for resource in vtf2.resources:
        assert _squash_xml(mivot_block.content) == _squash_xml(
            resource.mivot_block.content
        )
        assert len(resource.tables) == 1


def test_write_no_mivot():
    """
    Build a VOTable, put an empty MIVOT block in the first resource, checks it can be retrieved
    as well as the following table
    """

    vtf = VOTableFile()
    mivot_resource = Resource()
    mivot_resource.type = "meta"
    # pack the meta resource in a top level resource
    r1 = Resource()
    r1.type = "results"
    r1.resources.append(mivot_resource)
    t1 = tree.TableElement(vtf)
    t1.name = "t1"
    r1.tables.append(t1)
    vtf.resources.append(r1)
    # Push the VOTable in an IOSTream (emulates a disk saving)
    buff = io.BytesIO()
    vtf.to_xml(buff)

    # Read the IOStream (emulates a disk readout)
    buff.seek(0)
    vtf2 = parse(buff)
    assert len(vtf2.resources) == 1
    for resource in vtf2.resources:
        assert (
            _squash_xml(resource.mivot_block.content)
            == "<VODMLxmlns=http://www.ivoa.net/xml/mivot><REPORTstatus=KO>NoMivotblock</REPORT></VODML>"
        )
        assert len(resource.tables) == 1


def test_mivot_write_after_resource():
    """
    Build a VOTable, put a MIVOT block in the first resource after another meta resource,
    checks it can be retrieved as well as the following table
    """

    mivot_block = MivotBlock(
        """
    <VODML xmlns="http://www.ivoa.net/xml/mivot" >
      <REPORT status="OK">Unit test mivot block1</REPORT>
      <GLOBALS>
      </GLOBALS>
    </VODML>
    """
    )
    vtf = VOTableFile()
    mivot_resource = Resource()
    mivot_resource.type = "meta"
    mivot_resource.mivot_block = mivot_block
    # pack the meta resource in a top level resource
    r1 = Resource()
    r1.type = "results"
    i1 = tree.Info(name="test_name", value="test_value")
    r1.infos.append(i1)
    meta_resource = Resource()
    meta_resource.type = "meta"
    r1.resources.append(meta_resource)
    r1.resources.append(mivot_resource)
    t1 = tree.TableElement(vtf)
    t1.name = "t1"
    r1.tables.append(t1)
    vtf.resources.append(r1)
    # Push the VOTable in an IOSTream (emulates a disk saving)
    buff = io.BytesIO()
    vtf.to_xml(buff)

    # Read the IOStream (emulates a disk readout)
    buff.seek(0)
    vtf2 = parse(buff)
    assert len(vtf2.resources) == 1
    for resource in vtf2.resources:
        assert _squash_xml(mivot_block.content) == _squash_xml(
            resource.mivot_block.content
        )
        assert len(resource.tables) == 1


def test_mivot_forbidden_write():
    """
    Build a meta resource containing a MIVOT block,
    build the dummy MIVOT block first.
    """
    mivot_block = MivotBlock(
        """
    <VODML xmlns="http://www.ivoa.net/xml/mivot" >
       <REPORT status="KO">Unit test mivot block1</REPORT>
       <GLOBALS/>
    </VODML>
    """
    )
    # package the MIVOT block in the resource
    mivot_resource = Resource()
    mivot_resource.type = "results"

    with pytest.raises(E26):
        # A MIVOT block must be with "type=meta"
        mivot_resource.mivot_block = mivot_block


def test_mivot_order(tmp_path):
    """
    Build a VOTable with 2 resources containing MivotBlock, parse it, and write it in a file.
    Then compare it with another file to see if the order of the elements in a resource is respected,
    in particular the MivotBlock which should be before the tables.
    """

    mivot_block = MivotBlock(
        """
    <VODML xmlns="http://www.ivoa.net/xml/mivot" >
    </VODML>
    """
    )
    vtf = VOTableFile()

    mivot_resource = Resource()
    mivot_resource.type = "meta"
    mivot_resource.mivot_block = mivot_block

    mivot_resource2 = Resource()
    mivot_resource2.type = "meta"
    mivot_resource2.mivot_block = mivot_block

    # R1 : 2 mivot_block, 2 tables, 1 description, 1 info, 1 CooSys
    r1 = Resource()
    r1.type = "results"

    t1 = tree.TableElement(vtf)
    t1.name = "t1"
    t2 = tree.TableElement(vtf)
    t2.name = "t2"

    r1.tables.append(t1)
    r1.tables.append(t2)

    r1.resources.append(mivot_resource)
    r1.resources.append(mivot_resource2)

    cs = tree.CooSys(ID="_XYZ", system="ICRS")
    r1.coordinate_systems.append(cs)
    i1 = tree.Info(name="test_name", value="test_value")
    r1.infos.append(i1)

    vtf.resources.append(r1)

    # R2 : 1 resource "results", 1 mivot_block and 1 table
    r2 = Resource()
    r2.type = "results"

    r3 = Resource()
    r3.type = "results"

    t3 = tree.TableElement(vtf)
    t3.name = "t3"
    r2.tables.append(t3)
    r2.resources.append(mivot_resource)
    r2.resources.append(r3)

    vtf.resources.append(r2)

    # Push the VOTable in an IOSTream (emulates a disk saving)
    buff = io.BytesIO()
    vtf.to_xml(buff)

    # Read the IOStream (emulates a disk readout)
    buff.seek(0)
    vtf2 = parse(buff)

    vpath = get_pkg_data_filename("data/test.order.xml")
    vpath_out = str(tmp_path / "test.order.out.xml")
    vtf2.to_xml(vpath_out)

    # We want to remove the xml header from the VOTable
    with open(vpath_out) as file:
        lines = file.readlines()
    # The xml header is on 2 lines (line 2 and 3)
    del lines[1]
    del lines[1]

    with open(vpath_out, "w") as file:
        file.writelines(lines)

    assert filecmp.cmp(vpath, vpath_out)


def test_version_checks_from_api():
    """Test that a VOTable can be constructed from APIs without triggering
    version check failure.
    """
    votable = VOTableFile()  # v1.4 by default
    resource = Resource()
    votable.resources.append(resource)
    table0 = tree.TableElement(votable)  # Will now get the version check config.
    resource.tables.append(table0)
    # Not allowed in older VOTable formats.
    votable.set_all_tables_format("binary2")

    with pytest.raises(W37):
        votable.set_all_tables_format("unknown")

    votable = VOTableFile(version="1.1")
    resource = Resource()
    votable.resources.append(resource)
    table0 = tree.TableElement(votable)
    resource.tables.append(table0)
    with pytest.raises(W37):
        votable.set_all_tables_format("binary2")

    with pytest.raises(NotImplementedError):
        votable.set_all_tables_format("fits")

    # Check TimeSys can be created (requires version config check).
    # Force exception to make it easier to test.
    config = {"verify": "exception"}
    votable = VOTableFile(version="1.4", config=config)
    _ = tree.TimeSys("REF", timescale="UTC", config=votable.config)

    votable = VOTableFile(version="1.1", config=config)
    with pytest.raises(W54):
        tree.TimeSys("REF", timescale="UTC", config=votable.config)

    # CooSys was deprecated only in v1.2 and refposition only allowed in 1.5.
    votable = VOTableFile(version="1.2", config=config)
    with pytest.raises(W27):
        tree.CooSys("REF", system="ICRS", config=votable.config)

    votable = VOTableFile(version="1.4", config=config)
    with pytest.raises(W57):
        tree.CooSys(
            "REF", system="ICRS", refposition="Something", config=votable.config
        )

    votable = VOTableFile(version="1.5", config=config)
    _ = tree.CooSys(
        "REF", system="ICRS", refposition="Something", config=votable.config
    )

    # Param construction will warn less if v1.0 and no name.
    # Must use internal API to force 1.0 version.
    votable._version = "1.0"
    votable._config.update(votable._get_version_checks())
    with pytest.warns(W47):
        tree.Param(votable, ID="REF")

    votable.version = "1.3"
    with pytest.warns((W47, W15)):
        tree.Param(votable, ID="REF", datatype="int")
